"""This file contains tests for creating overflow graph. ie,
graph with delta flows, current_flows - flows_after_closing_line"""

import configparser
import networkx as nx
from alphaDeesp.core.printer import Printer
from alphaDeesp.core.network import Network
from alphaDeesp.core.pypownet.PypownetSimulation import PypownetSimulation
from alphaDeesp.core.pypownet.PypownetObservationLoader import PypownetObservationLoader
import numpy as np

def build_sim(ltc):
    config = configparser.ConfigParser()
    config.read("./alphaDeesp/tests/ressources_for_tests/config_for_tests.ini")
    param_folder = "./alphaDeesp/ressources/parameters/default14_static"
    timestep = 0

    loader = PypownetObservationLoader(param_folder)
    env, obs, action_space = loader.get_observation(timestep)
    sim = PypownetSimulation(env, obs, action_space, param_options=config["DEFAULT"], debug=False,
                             ltc=[ltc])
    return sim

def test_powerflow_graph():
    custom_layout = [(-280, -81), (-100, -270), (366, -270), (366, -54), (-64, -54), (-64, 54), (366, 0),
                     (438, 0), (326, 54), (222, 108), (79, 162), (-152, 270), (-64, 270), (222, 216),
                     (-280, -151), (-100, -340), (366, -340), (390, -110), (-14, -104), (-184, 54), (400, -80),
                     (438, 100), (326, 140), (200, 8), (79, 12), (-152, 170), (-70, 200), (222, 200)]
    config = configparser.ConfigParser()
    config.read("./alphaDeesp/tests/ressources_for_tests/config_for_tests.ini")
    param_folder = "./alphaDeesp/ressources/parameters/default14_static"

    sim = build_sim(ltc = 9)

    g_pow = sim.build_powerflow_graph(sim.obs)

    path_to_saved_graph = "./alphaDeesp/tests/ressources_for_tests/saved_graphs/gpow_geo_save_2019-08-05_18-26_0.dot"
    saved_g = nx.drawing.nx_pydot.read_dot(path_to_saved_graph)

    for e1, e2 in zip(saved_g.edges(data="label"), g_pow.edges(data="label")):
        assert (float(e1[2]) == float(e2[2]))

    print("g_over and saved_g are isomorphic: ", nx.is_isomorphic(g_pow, saved_g))
    assert (nx.is_isomorphic(g_pow, saved_g))

def test_overflow_grid():
    """This function, given the input folder in test/path,
    it computes the overflow graph and compares it with the saved graph: saved_overflow_graph.dot"""
    custom_layout = [(-280, -81), (-100, -270), (366, -270), (366, -54), (-64, -54), (-64, 54), (366, 0),
                     (438, 0), (326, 54), (222, 108), (79, 162), (-152, 270), (-64, 270), (222, 216),
                     (-280, -151), (-100, -340), (366, -340), (390, -110), (-14, -104), (-184, 54), (400, -80),
                     (438, 100), (326, 140), (200, 8), (79, 12), (-152, 170), (-70, 200), (222, 200)]

    sim = build_sim(ltc = 9)

    _grid = sim.environment.game.grid
    _current_observation = sim.obs

    sim.load_from_observation(_current_observation, [9])

    df_of_g = sim.get_dataframe()
    g_over = sim.build_graph_from_data_frame([9])

    path_to_saved_graph = "./alphaDeesp/tests/ressources_for_tests/saved_graphs/g_overflow_print_geo_2020-09-03_12-20.dot"
        #"./alphaDeesp/tests/ressources_for_tests/saved_graphs/overflow_graph_example_geo_2019-11-06_16-40_0_.dot"

    saved_g = nx.drawing.nx_pydot.read_dot(path_to_saved_graph)


    for e1 in saved_g.edges(data="label"):
        for e2 in g_over.edges(data="label"):
            if int(e1[0]) == int(e2[0]) and int(e1[1]) == int(e2[1]):
                saved_flow = float(e1[2][1: -1])
                current_flow = float(e2[2])
                assert (saved_flow == current_flow)

    print("g_over and saved_g are isomorphic: ", nx.is_isomorphic(g_over, saved_g))
    #assert (nx.is_isomorphic(g_over, saved_g))


def test_constrained_path():
    """Given the input folder : default_14_static,
    we must get constrained path : cpath = 4, 5, 6"""


def test_new_flows_after_network_cut():
    """After closing one electric line, the flows should change on the network."""
    custom_layout = [(-280, -81), (-100, -270), (366, -270), (366, -54), (-64, -54), (-64, 54), (366, 0),
                     (438, 0), (326, 54), (222, 108), (79, 162), (-152, 270), (-64, 270), (222, 216),
                     (-280, -151), (-100, -340), (366, -340), (390, -110), (-14, -104), (-184, 54), (400, -80),
                     (438, 100), (326, 140), (200, 8), (79, 12), (-152, 170), (-70, 200), (222, 200)]

    print("Test new flows")
    sim = build_sim(ltc = 9)
    print("flows = ", sim.obs.active_flows_origin.astype(int))

    new_flows = sim.cut_lines_and_recomputes_flows([9])

    print("diff flows = ")
    print((new_flows - sim.obs.active_flows_origin).astype(int))
    result = (new_flows - sim.obs.active_flows_origin).astype(int)
    print(result != 0.0)
    print("there are {} non zero elements".format(np.count_nonzero(result)))

    # only the line that got cut should be at zero.
    assert (np.count_nonzero(result) == 19)

    # g_over df_of_g = sim.buildgra
    #
    # printer = Printer()
    # printer.display_geo(g_over, custom_layout, name="test_graph")


def extract_topology():
    """This function takes an observation and tests the function: extract_topo_from_obs"""
    print("Test extract_topology")
    sim = build_sim(ltc = 9)
    print("flows = ", sim.obs.active_flows_origin.astype(int))


def test_isgraph_connected():
    """This function test if a graph is connexe"""
    # Three different types of graph for tests
    # ============ Graph g1 ============
    #        5
    #         \>
    # 1->-2->-3->-4
    #      \>
    #      6
    # printer = Printer()

    g1 = nx.DiGraph()
    g1.add_edges_from([(1, 2), (2, 3), (2, 6), (3, 4), (5, 3)])
    # printer.display_geo(g1, name="test_graph_connected")
    print("graph g1 is connected: ", nx.is_weakly_connected(g1))
    assert (nx.is_weakly_connected(g1))
    g1.remove_edge(2, 3)
    print("After removing edge graph g1 is not_connected: ", nx.is_weakly_connected(g1))
    assert (not nx.is_weakly_connected(g1))
    # printer.display_geo(g1, name="test_graph")


def test_detailed_graph_with_node_0_split_in_two():
    """
    From an observation, avec using load2()
    basic graph with zero nodes split in two, has 14 nodes
    with node 0 split in two, we should have 15 nodes
    :return:
    """
    sim = build_sim(ltc = 9)

    _grid = sim.environment.game.grid
    _current_observation = sim.obs

    # new_configuration = [[1, 1, 0], [0, 0, 0, 0, 1]]
    new_configuration = [[0, 1, 0]]
    # node_id = [1, 5]
    node_id = [1]
    # new_configuration = [[1, 1, 1], [1, 1, 1, 1, 1]]
    # node_id = [1, 5]
    _current_observation = sim.change_nodes_configurations(new_configuration, node_id)
    # print(type(_current_observation))
    print(_current_observation)

    sim.load_from_observation(_current_observation, [9])
    # #################################################### PUT THE ABOVE IN A FIXTURE PYTEST ?

    # g_over, df_of_g = sim.build_graph_from_data_frame([9])
    # g_over_detailed = sim.build_detailed_graph_from_internal_structure([9])

    network = Network(sim.substations_elements)
    print("There are {} graphical nodes ".format(network.get_graphical_number_of_nodes()))
    assert (network.get_graphical_number_of_nodes() == 15)


def test_from_dict_substations_elements_14_nodes_111_on_node_0_and_11111_on_node_5_should_have_16_nodes():
    sim = build_sim(ltc = 9)

    _current_observation = sim.obs

    new_configuration = [[1, 1, 1], [1, 1, 1, 1, 1]]
    node_id = [1, 5]
    _current_observation = sim.change_nodes_configurations(new_configuration, node_id)

    sim.load_from_observation(_current_observation, [9])

    network = Network(sim.substations_elements)
    assert (network.get_graphical_number_of_nodes() == 16)


def test_from_dict_substations_elements_14_nodes_111_on_node_0_and_11000_on_node_5_should_have_16_nodes():
    sim = build_sim(ltc = 9)

    _current_observation = sim.obs

    new_configuration = [[1, 1, 1], [1, 1, 0, 0, 0]]
    node_id = [1, 5]
    _current_observation = sim.change_nodes_configurations(new_configuration, node_id)

    sim.load_from_observation(_current_observation, [9])

    network = Network(sim.substations_elements)
    print("There are {} graphical nodes ".format(network.get_graphical_number_of_nodes()))
    assert (network.get_graphical_number_of_nodes() == 16)


def test_from_dict_substations_elements_14_nodes_110_on_node_0_and_11000_on_node_5_should_have_16_nodes():
    # read config file and parameters folder for Pypownet
    config = configparser.ConfigParser()
    config.read("./alphaDeesp/tests/ressources_for_tests/config_for_tests.ini")
    param_folder = "./alphaDeesp/ressources/parameters/default14_static"

    # run Pypownet
    sim = build_sim(ltc = 9)

    # retrieve Topology as an Observation object from Pypownet
    _current_observation = sim.obs

    # new configuration creation
    new_configuration = [[1, 1, 0], [1, 1, 0, 0, 0]]
    node_ids = [1, 5]

    # change network Topology and retrieve an Obersation object from Pypownet
    _current_observation = sim.change_nodes_configurations(new_configuration, node_ids)

    sim.load_from_observation(_current_observation, [9])
    #
    print("sim.substations_elements =", sim.substations_elements)
    network = Network(sim.substations_elements)
    assert (network.get_graphical_number_of_nodes() == 16)

# test_powerflow_graph()
# test_overflow_grid()
# test_new_flows_after_network_cut()
# test_isgraph_connected()
# test_detailed_graph_with_node_0_split_in_two()
# test_integration_dataframe_results_with_line_9_cut()
# test_save_red_dataframe()
# test_round_random_tests()



#!/usr/bin/python3
# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of ExpertOp4Grid, an expert system approach to solve flow congestions in power grids

from make_evaluation_env import make_grid2op_evaluation_env
from make_training_env import make_grid2op_training_env
from load_evaluation_data import list_all_chronics, get_first_obs_on_chronic
from datetime import datetime
from data_utils import StateInfo
import sys

import configparser
import numpy as np
import os
from alphaDeesp.core.grid2op.Grid2opObservationLoader import Grid2opObservationLoader
from alphaDeesp.core.grid2op.Grid2opSimulation import Grid2opSimulation
from alphaDeesp.core.graphsAndPaths import OverFlowGraph,Structured_Overload_Distribution_Graph
import networkx as nx
import pypowsybl as pp
import pandas as pd
from load_training_data import aux_prevent_asset_reconnection,load_interesting_lines,DELETED_LINE_NAME


import glob
import shutil
import json
from packaging.version import Version as version_packaging

from packaging.version import Version as version_packaging
from importlib.metadata import version
EXOP_MIN_VERSION = version_packaging("0.2.6")
if version_packaging(version("expertop4grid")) < EXOP_MIN_VERSION:
    raise RuntimeError(f"Incompatible version found for expertOp4Grid, make sureit is >= {EXOP_MIN_VERSION}")

param_options_expertOp={
    # 0.2 is 20 percent of the max overload flow
    "ThresholdReportOfLine": 0.05,#0.2,#0.05,#
    # 10 percent de la surcharge max
    "ThersholdMinPowerOfLoop": 0.1,
    # If at least a loop is detected, only keep the ones with a flow  of at least 25 percent the biggest one
    "ratioToKeepLoop": 0.25,
    # Ratio percentage for reconsidering the flow direction
    "ratioToReconsiderFlowDirection": 0.75,
    # max unused lines
    "maxUnusedLines": 3,
    # number of simulated topologies node at the final simulation step
    "totalnumberofsimulatedtopos": 30,
    # number of simulated topologies per node at the final simulation step
    "numberofsimulatedtopospernode": 10
    }


def identify_overload_lines_to_keep_overflow_graph_connected(obs_simu, lines_overloaded_ids):
    """
    This function identifies which overloaded lines to keep in order to maintain a connected overflow graph.

    Parameters:
    obs_simu (Grid2opSimulation): The simulation object containing the current state of the grid.
    lines_overloaded_ids (list): List of IDs of the overloaded lines.

    Returns:
    list: List of overloaded line IDs to keep in the overflow graph.
    """
    # Get the energy graph from the simulation object
    obs_graph = obs_simu.get_energy_graph()

    # Find the maximum rho value among all lines
    max_rho = max(obs_simu.rho)

    # Identify edges that are disconnected (rho value is 0)
    edges_disconnected = [edge for edge, rho in nx.get_edge_attributes(obs_graph, 'a_or').items() if np.round(rho, 3) == 0]

    # Identify the edge with the maximum rho value
    recover_max_overload_edge = [edge for edge, rho in nx.get_edge_attributes(obs_graph, 'rho').items() if rho == max(obs_simu.rho)]

    # Identify other overloaded edges
    recover_other_overload_edge = [edge for edge, rho in nx.get_edge_attributes(obs_graph, 'rho').items() for line_id in lines_overloaded_ids if rho == obs_simu.rho[line_id] and rho != max_rho]

    # Create a graph from the energy graph
    obs_graph = nx.Graph(obs_graph)

    # Remove disconnected edges from the graph
    obs_graph.remove_edges_from(edges_disconnected)

    # Calculate the number of connected components initially
    n_connected_comp_init = len([c for c in nx.connected_components(obs_graph)])

    # Remove the edge with the maximum overload from the graph
    obs_graph.remove_edges_from(recover_max_overload_edge)

    # Calculate the number of connected components after removing the max overload edge
    n_connected_comp_max_overload = len([c for c in nx.connected_components(obs_graph)])

    # Remove other overloaded edges from the graph
    obs_graph.remove_edges_from(recover_other_overload_edge)

    # Calculate the number of connected components after removing all overload edges
    n_connected_comp_all_overload = len([c for c in nx.connected_components(obs_graph)])

    # Determine which overloaded lines to keep based on the number of connected components
    if n_connected_comp_init == n_connected_comp_all_overload:
        # If removing all overloads does not change the number of connected components, keep all overloaded lines
        lines_overloaded_ids_to_keep = lines_overloaded_ids
    elif n_connected_comp_max_overload == n_connected_comp_init:
        # If removing the max overload edge does not change the number of connected components, keep only the max overload line
        lines_overloaded_ids_to_keep = [line_id for line_id in lines_overloaded_ids if obs_simu.rho[line_id] == max_rho]
        print(f"we reduce the problem by focusing on the deepest overload {obs_simu.name_line[lines_overloaded_ids_to_keep[0]]} as considering all overloads cut break the network appart")
    else:
        # If removing any overload edge changes the number of connected components, it is not solvable without load shedding
        lines_overloaded_ids_to_keep = None

    return lines_overloaded_ids_to_keep

def inhibit_swapped_flows(df_of_g):
    """
    This function readjusts the flow direction for swapped flows in the overflow graph, inhibiting this legacy behavior from expertOp4grid.

    Parameters:
    df_of_g (pandas.DataFrame): DataFrame containing the overflow graph data.

    Returns:
    pandas.DataFrame: Updated DataFrame with adjusted flow directions for swapped flows.
    """
    # Reverse delta_flows value for swapped flows
    df_of_g.loc[df_of_g.new_flows_swapped, "delta_flows"] = -df_of_g[df_of_g.new_flows_swapped]["delta_flows"]

    # Swap idx_or and idx_ex for swapped flows
    idx_or = df_of_g[df_of_g.new_flows_swapped]["idx_or"]
    df_of_g.loc[df_of_g.new_flows_swapped, "idx_or"] = df_of_g[df_of_g.new_flows_swapped]["idx_ex"]
    df_of_g.loc[df_of_g.new_flows_swapped, "idx_ex"] = idx_or

    return df_of_g

def build_overflow_graph(obs_overloaded,action_space,observation_space,overloaded_line_ids,non_connected_reconnectable_lines,lines_non_reconnectable,param_options,timestep,do_consolidate_graph=True,inhibit_swapped_flow_reversion=True):
    """
    This function builds an overflow graph based on the given overloaded observation and parameters.
    sim = Grid2opSimulation(obs_overloaded, action_space, observation_space, param_options=param_options, debug=False,
    Parameters:
    obs_overloaded (Grid2opObservation): The observation containing overloaded lines.
    action_space (Grid2opActionSpace): The action space for the environment.
    observation_space (Grid2opObservationSpace): The observation space for the environment.
    overloaded_line_ids (list): List of IDs of the overloaded lines.
    non_connected_reconnectable_lines (list): List of reconnectable lines.
    lines_non_reconnectable (list): List of non reconnectable lines.
    param_options (dict): Dictionary of parameters for the simulation.
    timestep (int): The timestep for the simulation.
    inhibit_swapped_flow_reversion (bool): Cancel the swapped edge direction for swapped flows (possibly not needed anymore given the new consolidate graph functions)

    Returns:
    tuple: A tuple containing five objects:
           - df_of_g: pandas dataframe with changes in flows for each lines after disconnecting overflows
           - overflow_sim (Grid2opSimulation): The simulation object.
           - g_overflow (OverFlowGraph): The overflow graph.
           - real_hubs: list of hubs in g_distribution_graph
           - g_distribution_graph (Structured_Overload_Distribution_Graph): The structured overload distribution graph.
    """
    overflow_sim = Grid2opSimulation(obs_overloaded, action_space, observation_space, param_options=param_options, debug=False,
                            ltc=overloaded_line_ids, plot=True, simu_step=timestep)

    df_of_g = overflow_sim.get_dataframe()

    #add line names
    df_of_g["line_name"] = obs_overloaded.name_line

    #inhibit swap flows, reverse delta_flows value for swapped flows
    if inhibit_swapped_flow_reversion:
        df_of_g=inhibit_swapped_flows(df_of_g)

    g_overflow = OverFlowGraph(overflow_sim.topo, overloaded_line_ids, df_of_g, float_precision="%.0f")

    mapping = {i: name for i, name in enumerate(obs_overloaded.name_sub)}
    g_overflow.g = nx.relabel_nodes(g_overflow.g, mapping, copy=True)

    g_distribution_graph = Structured_Overload_Distribution_Graph(g_overflow.g)

    if len(g_distribution_graph.g_only_red_components.nodes)!=0 and do_consolidate_graph:
        g_overflow.consolidate_graph(g_distribution_graph)
    #
        g_distribution_graph = Structured_Overload_Distribution_Graph(g_overflow.g)

    #get real hubs before we add disconected but reconnectable lines
    real_hubs = g_distribution_graph.get_hubs()

    #add disconected but reconnectable lines
    # TO DO - reminder: uncomment!!
    for i in range(2):#need two iterations to identify CHALOL31LOUHA reconnectable path under contingency "BEON L31CPVAN" at timestep 1 on chronic 28th august
        g_overflow.add_relevant_null_flow_lines_all_paths(g_distribution_graph, non_connected_lines=non_connected_reconnectable_lines,non_reconnectable_lines=lines_non_reconnectable)
        g_distribution_graph = Structured_Overload_Distribution_Graph(g_overflow.g)

    return df_of_g,overflow_sim,g_overflow,real_hubs,g_distribution_graph

def is_nodale_grid2op_action(act):

    #check if bus set happens for multiple objects at a substation
    subs, counts = np.unique(act._topo_vect_to_sub[(act._set_topo_vect >= 1)], return_counts=True)
    is_nodale_action = np.any(counts>=2)#if at least two objects at a given substation appear with a set bus
    concerned_subs=[]
    is_splitting_subs=[]
    if is_nodale_action:
        concerned_subs=[sub for i,sub in enumerate(subs) if counts[i]>=2]
        for sub in concerned_subs:
            bus_for_set=np.unique(act._set_topo_vect[act._topo_vect_to_sub==sub])
            is_splitting=len(bus_for_set[bus_for_set>=1])>=2
            is_splitting_subs.append(is_splitting)

    return is_nodale_action, concerned_subs, is_splitting_subs

def is_line_disconnection(grid2op_action):
    line_or_change = grid2op_action.line_or_change_bus
    line_ex_change = grid2op_action.line_ex_change_bus
    line_change_status = grid2op_action.line_change_status

    line_or_set = grid2op_action.line_or_set_bus
    line_ex_set = grid2op_action.line_ex_set_bus
    line_set_status = grid2op_action.line_set_status

    is_line_deconnection=False
    if np.any(line_change_status!=0) or np.any(line_or_change!=0) or np.any(line_ex_change!=0):
        print("WARNING: line_change_status is not supported in this is_line_deconnection function ")

    if np.any(line_or_set==-1) or np.any(line_ex_set==-1) or np.any(line_set_status==-1):
        is_line_deconnection=True

    return is_line_deconnection


def is_line_reconnection(grid2op_action):
    line_or_change = grid2op_action.line_or_change_bus
    line_ex_change = grid2op_action.line_ex_change_bus
    line_change_status = grid2op_action.line_change_status

    line_or_set = grid2op_action.line_or_set_bus
    line_ex_set = grid2op_action.line_ex_set_bus
    line_set_status = grid2op_action.line_set_status

    is_line_reconnection=False
    if np.any(line_change_status!=0) or np.any(line_or_change!=0) or np.any(line_ex_change!=0):
        print("WARNING: line_change_status is not supported in this is_line_reconnection function ")

    if np.any(line_or_set*line_ex_set == 1) or np.any(line_set_status == 1):
        is_line_reconnection = True

    return is_line_reconnection

def is_load_disconnection(grid2op_action):
    load_change_bus = grid2op_action.load_change_bus
    load_set_bus = grid2op_action.load_set_bus

    is_load_disconnection=False
    if np.any(load_change_bus!=0):
        print("WARNING: load_change_bus is not supported in this is_load_disconnection function ")

    if np.any(load_set_bus == -1):
        is_load_disconnection = True

    return is_load_disconnection

def identify_grid2op_action_type(grid2op_action):
    """
    This function identifies the type of action based on the provided grid2op action object.

    Parameters:
    grid2op_action (grid2op.Action): The action object from the Grid2Op library.

    Returns:
    str: A string representing the type of action. Possible values are:
         - "open_coupling"
         - "close_coupling"
         - "open_line"
         - "close_line"
         - "open_load"
         - "close_load"
         - "open_line_load"
         - "close_line_load"
    """
    is_nodale_action, concerned_subs, is_splitting_subs=is_nodale_grid2op_action(grid2op_action)
    if is_nodale_action:
        if any(is_splitting_subs):
            return "open_coupling"
        else:
            return "close_coupling"
    else:
        is_line_disco=is_line_disconnection(grid2op_action)
        is_line_reco=is_line_reconnection(grid2op_action)
        is_load_disco=is_load_disconnection(grid2op_action)

        if is_line_disco:
            if is_load_disco:
                return "open_line_load"
            else:
                return "open_line"
        if is_line_reco:
            return "close_line"
        if is_load_disco:
            return "open_load"

    return "unknown"


def identify_action_type(actions_desc, by_description=True,grid2op_action_space=None):
    """
    This function identifies the type of action based on the provided action description.

    Parameters:
    actions_desc (dict): A dictionary containing the action description.
                         Expected keys include "description_unitaire" and "content".
    by_description (bool): A flag indicating whether to identify the action type based on the description.
                          If False, the function will identify the action type based on the content directly.

    Returns:
    str: A string representing the type of action. Possible values are:
         - "open_coupling"
         - "close_coupling"
         - "open_line"
         - "close_line"
         - "open_load"
         - "close_load"
         - "open_line_load"
         - "close_line_load"
    """
    type = None
    if by_description:
        # Extract the set_bus dictionary from the action description
        dict_action = actions_desc["content"]["set_bus"]

        # Check if there are any loads or lines involved in the action
        has_load = len(dict_action["loads_id"]) != 0
        has_line = (len(dict_action["lines_or_id"]) != 0) or (len(dict_action["lines_ex_id"]) != 0)

        # Get the description of the action
        description = actions_desc["description_unitaire"]

        # Determine the action type based on the description
        if ("COUPL" in description or "TRO." in description) and "Ouverture" in description:
            type = "open_coupling"
        elif ("COUPL" in description or "TRO." in description) and "Fermeture" in description:
            type = "close_coupling"
        elif "Ouverture" in description:
            if has_load and has_line:
                type = "open_line_load"
            elif has_line:
                type = "open_line"
            else:
                type = "open_load"
        elif "Fermeture" in description:
            if has_load and has_line:
                type = "close_line_load"
            elif has_line:
                type = "close_line"
            else:
                type = "close_load"
    else:
        # If by_description is False, the function does not process the action type based on the content directly
        grid2op_action_str=actions_desc["content"]
        grid2op_action=grid2op_action_space(grid2op_action_str)
        type=identify_grid2op_action_type(grid2op_action)

    return type

def localize_line_action(lines, lines_constrained_path, lines_dispatch):
    """
    This function determines the localization of a set of lines within the constrained path or dispatch path.

    Parameters:
    lines (list): List of line IDs to be localized.
    lines_constrained_path (list): List of line IDs that are part of the constrained path.
    lines_dispatch (list): List of line IDs that are part of the dispatch path.

    Returns:
    str: A string indicating the localization of the lines. Possible values are:
         - "out_of_graph": The lines are not part of the constrained path or dispatch path.
         - "constrained_path": The lines are part of the constrained path.
         - "dispatch_path": The lines are part of the dispatch path.
    """
    localization = "out_of_graph"

    lines_intersect_constrained_path = set(lines).intersection(set(lines_constrained_path))
    lines_intersect_dispatch_lines = set(lines).intersection(set(lines_dispatch))

    if len(lines_intersect_constrained_path) != 0:
        localization = "constrained_path"
    elif len(lines_intersect_dispatch_lines) != 0:
        localization = "dispatch_path"

    return localization

def localize_coupling_action(action_subs, hubs, nodes_constrained_path, nodes_dispatch_path):
    """
    This function determines the localization of a coupling action within the constrained path, dispatch path, or hubs.

    Parameters:
    action_subs (list): List of action sub-topologies to be localized.
    hubs (list): List of hub nodes in the graph.
    nodes_constrained_path (list): List of nodes that are part of the constrained path.
    nodes_dispatch_path (list): List of nodes that are part of the dispatch path.

    Returns:
    str: A string indicating the localization of the coupling action. Possible values are:
         - "out_of_graph": The action is not part of the constrained path, dispatch path, or hubs.
         - "hubs": The action is part of the hubs.
         - "constrained_path": The action is part of the constrained path.
         - "dispatch_path": The action is part of the dispatch path.
    """
    # Initialize the localization to "out_of_graph"
    localization = "out_of_graph"

    # Check if the action involves any hubs
    action_in_hubs = len(set(action_subs).intersection(set(hubs))) != 0

    # Check if the action involves any nodes in the constrained path
    action_in_constrained_path = len(set(action_subs).intersection(set(nodes_constrained_path))) != 0

    # Check if the action involves any nodes in the dispatch path
    action_in_dispatch_path = len(set(action_subs).intersection(set(nodes_dispatch_path))) != 0

    # Determine the localization based on the checks
    if action_in_hubs:
        localization = "hubs"
    elif action_in_constrained_path:
        localization = "constrained_path"
    elif action_in_dispatch_path:
        localization = "dispatch_path"

    return localization

def check_rules(action_type, localization, subs_topology):
    """
    This function checks if the given action violates any expert-defined rules based on its type, localization, and sub-topology.

    Parameters:
    action_type (str): The type of the action (e.g., "open_line", "close_coupling").
    localization (str): The localization of the action within the graph (e.g., "dispatch_path", "constrained_path").
    subs_topology (list): A list of sub-topologies involved in the action.

    Returns:
    tuple: A tuple containing two elements:
           - do_filter_action (bool): A flag indicating whether the action should be filtered out.
           - broken_rule (str): A string describing the rule that was broken, if any.
    """
    do_filter_action = False
    broken_rule = None

    if "load" not in action_type:  # We don't filter actions that disconnect loads for now
        # Check if the topology is in a single node
        is_topo_subs_one_node = np.all([len(set(sub_topo) - set([-1])) == 1 for sub_topo in subs_topology])

        # Expert rules to prevent
        out_of_graph = (localization == "out_of_graph")
        line_disconnection_dispatch_path = ("line" in action_type) and ("open" in action_type) and (localization == "dispatch_path")
        line_reconnection_constrained_path = ("line" in action_type) and ("close" in action_type) and (localization == "constrained_path")
        # Filter node splitting only if the sub-topology was initially one node
        node_splitting_dispatch_path = ("coupling" in action_type) and ("open" in action_type) and (localization == "dispatch_path") and is_topo_subs_one_node
        node_merging_constrained_path = ("coupling" in action_type) and ("close" in action_type) and (localization == "constrained_path")
        # Check if any rules are broken
        do_filter_action = out_of_graph or line_disconnection_dispatch_path or line_reconnection_constrained_path or node_splitting_dispatch_path or node_merging_constrained_path

        if out_of_graph:
            broken_rule = "No action out of the overflow graph"
        elif line_reconnection_constrained_path:
            broken_rule = "No line reconnection on constrained path"
        elif line_disconnection_dispatch_path:
            broken_rule = "No line disconnection on dispatch path"
        elif node_merging_constrained_path:
            broken_rule = "No node merging on constrained path"
        elif node_splitting_dispatch_path:
            broken_rule = "No node splitting on dispatch path"

    return do_filter_action, broken_rule

def verify_action(action_desc, hubs, lines_constrained_path, nodes_constrained_path, lines_dispatch, nodes_dispatch_path, subs_topology=[],by_description=True,grid2op_action_space=None):
    """
    This function verifies whether an action should be filtered based on expert-defined rules.

    Parameters:
    action_desc (dict): A dictionary containing the action description.
                         Expected keys include "description_unitaire", "content", and "VoltageLevelId".
    hubs (list): List of hub nodes in the graph.
    lines_constrained_path (list): List of line IDs that are part of the constrained path.
    nodes_constrained_path (list): List of nodes that are part of the constrained path.
    lines_dispatch (list): List of line IDs that are part of the dispatch path.
    nodes_dispatch_path (list): List of nodes that are part of the dispatch path.
    subs_topology (list): A list of sub-topologies involved in the action. Default is an empty list.

    Returns:
    tuple: A tuple containing two elements:
           - do_filter_action (bool): A flag indicating whether the action should be filtered out.
           - broken_rule (str): A string describing the rule that was broken, if any.
    """
    # in current implementation, we need "description_unitaire","content (with "set_bus") and "VoltageLevelId" fields in actions_desc
    # Identify the type of action based on the description
    action_type = identify_action_type(action_desc, by_description,grid2op_action_space)

    # Determine the localization of the action within the graph
    if "line" in action_type:
        grid2op_actions_set_bus = action_desc["content"]["set_bus"]
        lines = list(grid2op_actions_set_bus["lines_or_id"].keys()) + list(
            grid2op_actions_set_bus["lines_ex_id"].keys())
        localization = localize_line_action(lines, lines_constrained_path, lines_dispatch)
    else:
        action_subs = [action_desc["VoltageLevelId"]]
        localization = localize_coupling_action(action_subs, hubs, nodes_constrained_path, nodes_dispatch_path)

    # Check if the action violates any expert-defined rules
    do_filter_action, broken_rule = check_rules(action_type, localization, subs_topology)

    # If the action is filtered out and no specific rule is broken, print a check message
    if do_filter_action and broken_rule is None:
        print("check")

    return do_filter_action, broken_rule

def check_rho_reduction(obs, timestep, act_defaut, action, overload_ids,act_reco_maintenance, rho_tolerance=0.01):
    """
    This function checks if the action reduces the rho values of overloaded lines.
    is_rho_reduction=None
    Parameters:
    obs (Grid2opObservation): The initial observation of the grid.
    timestep (int): The timestep for the simulation.
    act_defaut (Grid2opAction): The default action to be applied.
    action (Grid2opAction): The action to be checked.
    overload_ids (list): List of IDs of the overloaded lines.
    rho_tolerance (float): Tolerance value to determine if a reduction in rho is significant. Default is 0.01.

    Returns:
    bool: True if the action results in a reduction of rho values for the overloaded lines, False otherwise.
    """
    # Initialize the flag to check if rho reduction is observed
    is_rho_reduction = None

    # Simulate the default action to get the initial rho values
    obs_defaut, reward, done, info = obs.simulate(act_defaut+act_reco_maintenance, time_step=timestep)
    rho_init = obs_defaut.rho[overload_ids]

    # Simulate the combined action (default action + additional action) to get the final rho values
    obs_simu_action, reward, done, info = obs.simulate(action + act_defaut+act_reco_maintenance, time_step=timestep)
    rho_final = obs_simu_action.rho[overload_ids]

    # Check if there are no exceptions in the simulation
    if len(info["exception"]) == 0:
        # Check if all final rho values are less than the initial rho values within the tolerance
        if np.all(rho_final + rho_tolerance < rho_init):
            is_rho_reduction = True
            print("We saw a reduction in rho from " + str(rho_init) + " to " + str(rho_final))
        else:
            is_rho_reduction = False

    return is_rho_reduction

def categorize_action_space(dict_action, hubs, lines_constrained_path, nodes_constrained_path, lines_dispatch,
                            nodes_dispatch_path, obs, timestep, defaut, action_space, overload_ids,lines_reco_maintenance, action_rule_preprocessing=True,by_description=True):
    """
    This function categorizes the action space based on expert-defined rules and verifies whether each action should be filtered out.
    actions_to_filter={}
    Parameters:
    dict_action (dict): A dictionary containing the action descriptions.
    hubs (list): List of hub nodes in the graph.
    lines_constrained_path (list): List of line IDs that are part of the constrained path.
    nodes_constrained_path (list): List of nodes that are part of the constrained path.
    lines_dispatch (list): List of line IDs that are part of the dispatch path.
    nodes_dispatch_path (list): List of nodes that are part of the dispatch path.
    obs (Grid2opObservation): The initial observation of the grid.
    timestep (int): The timestep for the simulation.
    defaut (str): The default line to be disconnected.
    action_space (Grid2opActionSpace): The action space for the environment.
    overload_ids (list): List of IDs of the overloaded lines.
    action_rule_preprocessing (bool): Flag to determine if action rule preprocessing should be applied. Default is True.

    Returns:
    tuple: A tuple containing two dictionaries:
           - actions_to_filter: Dictionary of actions that should be filtered out.
           - actions_unfiltered: Dictionary of actions that should not be filtered out.
    """
    actions_to_filter = {}
    actions_unfiltered = {}

    for action_id, action_desc in dict_action.items():
        # Determine the sub-topologies involved in the action
        action_subs = [action_desc["VoltageLevelId"]]
        subs_topology = [obs.sub_topology(np.where(obs.name_sub == sub_name)[0][0]) for sub_name in action_subs]

        # Verify if the action should be filtered based on expert-defined rules
        do_filter_action, broken_rule = verify_action(action_desc, hubs, lines_constrained_path, nodes_constrained_path,
                                                      lines_dispatch, nodes_dispatch_path, subs_topology,by_description,action_space)

        action_grid2op = action_desc["content"]

        if do_filter_action:
            # Create default and additional actions to simulate
            act_defaut = action_space({"set_bus": {"lines_ex_id": {defaut: -1}, "lines_or_id": {defaut: -1}}})
            action = action_space(action_grid2op)

            if action_rule_preprocessing:
                state = StateInfo()
                action = aux_prevent_asset_reconnection(obs, state, action)

            # Check if the action reduces the rho values of overloaded lines
            act_reco_maintenance = action_space(
                {"set_line_status": [(line_reco, 1) for line_reco in lines_reco_maintenance]})
            is_rho_reduction = check_rho_reduction(obs, timestep, act_defaut, action, overload_ids,act_reco_maintenance, rho_tolerance=0.01)

            if is_rho_reduction:
                print(action_desc["description_unitaire"])
                print(broken_rule)

            # Add the filtered action to the dictionary
            actions_to_filter[action_id] = {
                "description_unitaire": action_desc["description_unitaire"],
                "broken_rule": broken_rule,
                "is_rho_reduction": is_rho_reduction
            }
        else:
            # Add the unfiltered action to the dictionary
            actions_unfiltered[action_id] = {
                "description_unitaire": action_desc["description_unitaire"]
            }

    return actions_to_filter, actions_unfiltered

def make_overflow_graph_visualization(env_path, overflow_sim, g_overflow, obs_simu, save_folder, graph_file_name, lines_swapped,custom_layout=None, draw_only_significant_edges=True):
    """
    This function creates and saves a visualization of the overflow graph based on the given parameters.

    Parameters:
    env_path (str): The path to the environment folder.
    overflow_sim (Grid2opSimulation): The simulation object containing the overflow state.
    g_overflow (OverFlowGraph): The overflow graph object.
    obs_simu (Grid2opObservation): The observation containing the simulation state.
    save_folder (str): The folder where the visualization will be saved.
    graph_file_name (str): The name of the file to save the visualization.
    lines_swapped (list): List of line names that have swapped flows.
    draw_only_significant_edges (bool): Flag to determine if only significant edges should be drawn. Default is True.

    Returns:
    svg: The generated svg.
    """

    # Rescale factor for better layout visualization
    rescale_factor = 3#5  # Adjust this value to change the zoom level
    fontsize = 10
    node_thickness = 2
    shape_hub = "diamond"

    #####
    # Add voltage levels to the graph
    file_iidm = "grid.xiidm"
    network_file_path = os.path.join(env_path, file_iidm)
    n_zone = pp.network.load(network_file_path)
    df_volt = n_zone.get_voltage_levels()

    # Create a dictionary mapping sub-stations to their nominal voltages
    df_volt_dict = {sub: volt for sub, volt in zip(df_volt.index, df_volt.nominal_v)}

    # Define colors for different voltage levels
    voltage_colors = {
        400: "red", 225: "darkgreen", 90: "gold", 63: "purple", 20: "pink",
        24: "pink", 10: "pink", 15: "pink", 33: "pink"
    }

    # Set voltage level colors in the overflow graph
    g_overflow.set_voltage_level_color(df_volt_dict, voltage_colors)

    #####
    # Add node numbers to the graph
    number_nodal_dict = {
        sub_name: len(set(obs_simu.sub_topology(i)) - set([-1]))
        for i, sub_name in enumerate(obs_simu.name_sub)
    }
    g_overflow.set_electrical_node_number(number_nodal_dict)

    #######
    # Add hubs to the graph
    g_overflow.set_hubs_shape(g_distribution_graph.hubs, shape_hub=shape_hub)

    #######
    # Highlight limiting lines in the graph
    dict_significant_change_in_line_loading = {}

    ######
    # Highlight swapped flows in the graph
    g_overflow.highlight_swapped_flows(lines_swapped)

    # Identify assets to monitor based on rho values
    # But first check if load flow run in DC, because in that case we cannot really estimate the rho
    is_DC = obs_simu._obs_env._parameters.ENV_DC
    if is_DC:
        print("we ill not highlight the lines with high rho because the load flow is in DC and rho is not properly estimated in that case")
    else:
        ind_assets_to_monitor = np.where(overflow_sim.obs_linecut.rho >= 0.9)[0]

        lines_overloaded_ids = overflow_sim.ltc
        ind_assets_to_monitor = np.append(ind_assets_to_monitor, lines_overloaded_ids)
        line_loaded_of_interest = env.name_line[ind_assets_to_monitor]

        # Record significant changes in line loading before and after the simulation
        for ind_line, line_name in zip(ind_assets_to_monitor, line_loaded_of_interest):
            dict_significant_change_in_line_loading[line_name] = {
                "before": int(obs_simu.rho[ind_line] * 100),
                "after": int(overflow_sim.obs_linecut.rho[ind_line] * 100)
            }

        # Highlight significant line loading changes in the graph
        g_overflow.highlight_significant_line_loading(dict_significant_change_in_line_loading)

    #####
    # Generate SVG and save the visualization
    tmp_save_folder = os.path.join(save_folder, graph_file_name)
    svg = g_overflow.plot(
        custom_layout,
        save_folder=tmp_save_folder,
        fontsize=fontsize,
        without_gray_edges=draw_only_significant_edges,
        node_thickness=node_thickness,
        rescale_factor=rescale_factor
    )

    # Move and rename the saved PDF file
    pdf_files = glob.glob(f"{tmp_save_folder}/Base graph/*.pdf")
    file_path = os.path.join(save_folder, graph_file_name + ".pdf")
    shutil.move(pdf_files[0], file_path)
    shutil.rmtree(tmp_save_folder)

    print("Overflow graph visualization has been saved in: " + file_path)

    return svg

def save_to_json(data, output_file):
    """Saves extracted data to a json file."""
    with open(output_file, "w") as json_file:
        json.dump(data, json_file, indent=4)

def check_simu_overloads(obs,line_defaut,lines_overloaded_ids,lines_reco_maintenance):
    has_converged=False
    has_lost_load=False
    act_deco_overloads=env.action_space({"set_line_status": [(env.name_line[line_id], -1) for line_id in lines_overloaded_ids]})
    act_deco_defaut=env.action_space({"set_line_status": [(line_defaut, -1)]})
    act_reco_maintenance = env.action_space(
        {"set_line_status": [(line_reco, 1) for line_reco in lines_reco_maintenance]})
    obs_simu_overloads, reward, done, info = obs.simulate(act_deco_overloads+act_deco_defaut+act_reco_maintenance, time_step=timestep)
    if len(info["exception"])!=0:
        print(f"error in simulation of all overloads : {[env.name_line[line_id] for line_id in lines_overloaded_ids]} : {info['exception']}")
    elif obs_simu_overloads.load_p.sum()+1<obs_simu.load_p.sum():
        print(f"lost loads in simulation of all overloads : {[env.name_line[line_id] for line_id in lines_overloaded_ids]} : {obs_simu_overloads.load_p.sum()} < {obs_simu.load_p.sum()}")
        has_converged=True
        has_lost_load=True
    else:
        has_converged=True
        has_lost_load=False
    return has_converged,has_lost_load



if __name__ == "__main__":

    date = datetime(2024, 11, 25)#datetime(2024, 11, 25)#datetime(2024, 12, 9)#datetime(2024, 12, 2)#datetime(2024, 8, 28)  # we choose a date for the chronic
    timestep = 14#14#13#22 #1 # 36
    line_defaut = "BEON L31CPVAN"#"MAGNYY633"#"BEON L31CPVAN"##AISERL31RONCI, P.SAOL31RONCI, AISERL31MAGNY, BEON L31CPVAN, "FRON5L31LOUHA"
    env_folder="./"
    env_name = "env_dijon_v2_assistant"
    env_path=os.path.join(env_folder,env_name)

    #User parameters
    draw_only_significant_edges=True#True#False
    use_grid_layout=True #True use geo layout, otherwise hierarchical automatic layout
    check_with_action_description=True#if False it will use the grid2op action object directly for checking the rules, if True it only uses the human description field of the action
    use_evaluation_config = False  # if false, use training config
    use_dc_load_flow=False #if problem of convergence of AC load flow because of too many iterations when disconnecting overloads, use DC load flow for building the overflow graph. We will still run the first load flow with N-1 disconnection with AC load flow to detect the overloads
    do_consolidate_graph=True #use False in case too many actions have been unfortunately filtered out which might indicate that the graph consolidation heuristic might not have been so relevant in that case

    #additional inputs
    action_space_folder="action_space"
    file_action_space_desc="actions_repas_most_frequent_topologies_revised.json"
    file_path = os.path.join(action_space_folder, file_action_space_desc)

    #load actions
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"The file {file_path} does not exist.")

    with open(file_path, 'r') as file:
        dict_action = json.load(file)

    # make the environment
    # env = grid2op.make(env_name, backend=backend, n_busbar=6, param=p)
    if use_evaluation_config:
        env = make_grid2op_evaluation_env(env_folder, env_name)
    else:
        env = make_grid2op_training_env(env_folder, env_name)


    custom_layout=None
    if use_grid_layout:
        custom_layout = [env.grid_layout[sub] for sub in env.name_sub]

    chronics_name = list_all_chronics(env)
    print("chronics names are:")
    print(chronics_name)

    # we get the first observation for the chronic at the desired date
    path_chronic = [path for path in env.chronics_handler.real_data.subpaths if date.strftime('%Y%m%d') in path][0]
    obs = get_first_obs_on_chronic(date, env,path_thermal_limits=path_chronic)
    chronic_name = env.chronics_handler.get_name()



    # read non reconnectable lines
    lines_non_reconnectable = list(load_interesting_lines(path=path_chronic,file_name="non_reconnectable_lines.csv"))
    lines_should_not_reco_2024_and_beyond =DELETED_LINE_NAME

    #detect initially disconnected lines that might need to be reconnected
    maintenance_df = pd.DataFrame(env.chronics_handler.real_data.data.maintenance_handler.array, columns=obs.name_line)
    lines_in_maintenance_obs_start = list(maintenance_df.iloc[0][(maintenance_df.iloc[0])].index)
    reconnectabble_line_in_maintenance_at_start = list(
        set(lines_in_maintenance_obs_start) - set(lines_non_reconnectable))
    print(f"lines in maintenance at start: {lines_in_maintenance_obs_start}")
    print(f"lines in maintenance at start and reconnectable in scenario: {reconnectabble_line_in_maintenance_at_start}")

    #act to reconnect some initially disconnected lines ?
    do_reco_maintenance_at_t = ~maintenance_df[reconnectabble_line_in_maintenance_at_start].iloc[timestep]
    maintenance_to_reco_at_t = list(do_reco_maintenance_at_t[do_reco_maintenance_at_t].index)
    if len(maintenance_to_reco_at_t) != 0:
        print(f"reconnecting lines not in maintenance anymore {maintenance_to_reco_at_t} at timestep {timestep}")
    act_reco_maintenance = env.action_space(
        {"set_line_status": [(line_reco, 1) for line_reco in maintenance_to_reco_at_t]})

    # simulate contingency tp detect overloads
    act_deco_defaut = env.action_space({"set_line_status": [(line_defaut, -1)]})
    obs_simu, reward, done, info = obs.simulate(act_deco_defaut+act_reco_maintenance, time_step=timestep)
    if len(info["exception"])!=0:
        print(f"error in simulation of contingency : {line_defaut} : {info['exception']}")
        print("we cannot analyze the problem then and stop the computation here")
        sys.exit(0)#we stop here the execution of the script

    lines_overloaded_ids = [i for i, rho in enumerate(obs_simu.rho) if rho >= 1]

    non_connected_reconnectable_lines = [l_name for i,l_name in enumerate(env.name_line) if l_name not in lines_non_reconnectable+lines_should_not_reco_2024_and_beyond and not obs_simu.line_status[i]]

    ##########
    has_converged,has_lost_load=check_simu_overloads(obs,line_defaut, lines_overloaded_ids,maintenance_to_reco_at_t)

    #check if disconnecting overloads, either the largest or all, leads to disconnected components. Only keep the toughest one if that's the case
    lines_overloaded_ids_kept=identify_overload_lines_to_keep_overflow_graph_connected(obs_simu, lines_overloaded_ids)

    if lines_overloaded_ids_kept:
        #inspect other simu
        #act=act_deco_defaut
        #for ov_id in lines_overloaded_ids_kept[:1]:
        #    act+=env.action_space({"set_line_status": [(env.name_line[ov_id], -1)]})
#
        ###########
        if use_dc_load_flow:
            env_params = env.parameters
            env_params.ENV_DC = True
            if use_evaluation_config:
                env = make_grid2op_evaluation_env(env_folder, env_name,params=env_params)
            else:
                env = make_grid2op_training_env(env_folder, env_name,params=env_params)

            obs = get_first_obs_on_chronic(date, env)#reset the env and get first obs with DC load flow
            obs_simu, reward, done, info = obs.simulate(act_deco_defaut+act_reco_maintenance, time_step=timestep)
        ########"

        #check if diconnecting kept overloads still breaks the graph with loast loads
        if lines_overloaded_ids_kept!=lines_overloaded_ids:
            has_converged, has_lost_load = check_simu_overloads(obs,line_defaut, lines_overloaded_ids_kept,maintenance_to_reco_at_t)
            if not has_converged:
                print("this prevents us from building the overflow graph and the issue might not be solvable with topology except maybe with some line reconnections")
                sys.exit(0)  # we stop here the execution of the script

        inhibit_swapped_flow_reversion=True#Cancel the swapped edge direction for swapped flows (possibly not needed anymore given the new consolidate graph functions)
        df_of_g,overflow_sim,g_overflow,hubs, g_distribution_graph = build_overflow_graph(obs_simu, env.action_space, env.observation_space,
                                                                lines_overloaded_ids_kept, non_connected_reconnectable_lines,lines_non_reconnectable,
                                                                param_options_expertOp, timestep,do_consolidate_graph,inhibit_swapped_flow_reversion)

        ###########
        # make graph visualization and save
        graph_file_name = "Overflow_Graph_" + line_defaut +"_chronic_"+chronic_name+"_timestep_"+str(timestep)
        if use_grid_layout:
            graph_file_name += "_geo"
        else:
            graph_file_name += "_hierarchi"
        if draw_only_significant_edges:
            graph_file_name += "_only_signif_edges"
        else:
            graph_file_name += "_all_edges"
        if do_consolidate_graph:
            graph_file_name += "_consoli"
        else:
            graph_file_name += "_no_consoli"
        if use_dc_load_flow:
            graph_file_name += "_in_DC"
        save_folder="./Overflow_Graph"
        lines_swapped=list(df_of_g[df_of_g.new_flows_swapped].line_name)
        make_overflow_graph_visualization(env_path, overflow_sim, g_overflow, obs_simu,save_folder, graph_file_name,lines_swapped,custom_layout, draw_only_significant_edges)
        print("ok")

        ##########
        # get useful paths for action verification
        lines_constrained_path, nodes_constrained_path = g_distribution_graph.get_constrained_edges_nodes()
        #if lines_overloaded_ids_kept!=lines_overloaded_ids, check that all overloads are on constrained path
        if len(lines_overloaded_ids_kept)!=len(lines_overloaded_ids):
            missing_overload_constrained_path=set(env.name_line[lines_overloaded_ids])-set(lines_constrained_path)
            if len(missing_overload_constrained_path)!=0:
                print("WARNING: our overload graphs should be considered with caution as lines "+str(missing_overload_constrained_path)+" are not seen on the constrained path which would be more consistent")


        lines_dispatch, nodes_dispatch_path = g_distribution_graph.get_dispatch_edges_nodes()

        #########
        # check rules for each action
        actions_to_filter,actions_unfiltered=categorize_action_space(dict_action, hubs, lines_constrained_path, nodes_constrained_path, lines_dispatch,
                                nodes_dispatch_path, obs, timestep, line_defaut, env.action_space, lines_overloaded_ids,maintenance_to_reco_at_t,by_description=check_with_action_description)#here we check for all overloaded lines

        print("#################     NOW printing the actions that have been filtered out     #################")
        for action_id, action_content in actions_to_filter.items():
            print(action_content['description_unitaire'])
            print("    " + action_content['broken_rule'])
            print("\n")

        n_actions=len(dict_action.keys())
        n_actions_filtered=len(actions_to_filter.keys())
        n_actions_badly_filtered=len([id for id,act_filter_content in actions_to_filter.items() if act_filter_content["is_rho_reduction"]])

        print("#################     NOW printing the actions that have been UNFORTUNATELY filtered out     #################")
        for action_id, action_content in actions_to_filter.items():
            if action_content["is_rho_reduction"]:
                print(action_content['description_unitaire'])
                print("    " + action_content['broken_rule'])
                print("\n")

        print(str(n_actions_filtered)+" actions have been filtered out of "+str(n_actions))
        print(str(n_actions_badly_filtered) + " actions have been unfortunately filtered out of " + str(n_actions_filtered)+" since they showed a tendency to reduce a bit the overflow")


        save_to_json(actions_to_filter, "actions_to_filter_expert_rules.json")
        save_to_json(actions_unfiltered, "actions_unfiltered_expert_rules.json")

        if use_dc_load_flow:
            print("Warning: you have used the DC load flow, so results are more approximate")


    else:
        print("Overload breaks the grid apart, only load shedding actions are an option")


import configparser

from alphaDeesp.core.alphadeesp import AlphaDeesp
import pandas as pd
from pathlib import Path
import ast
import numpy as np
import pandas as pd
import time

from alphaDeesp.core.grid2op.Grid2opObservationLoader import Grid2opObservationLoader
from alphaDeesp.core.grid2op.Grid2opSimulation import Grid2opSimulation
from alphaDeesp.core.graphsAndPaths import OverFlowGraph, PowerFlowGraph


custom_layout = [(-280, -81), (-100, -270), (366, -270), (366, -54), (-64, -54), (-64, 54), (366, 0),
                 (438, 0), (326, 54), (222, 108), (79, 162), (-152, 270), (-64, 270), (222, 216),
                 (-280, -151), (-100, -340), (366, -340), (390, -110), (-14, -104), (-184, 54), (400, -80),
                 (438, 100), (326, 140), (200, 8), (79, 12), (-152, 170), (-70, 200), (222, 200)]


def build_sim(ltc, param_folder, config_file = "./alphaDeesp/tests/resources_for_tests_grid2op/config_for_tests.ini",
              chronic_scenario = None, timestep = 0,modified_thermal_Limit=None):
    config = configparser.ConfigParser()
    config.read(config_file)

    loader = Grid2opObservationLoader(param_folder)
    env, obs, action_space = loader.get_observation(timestep = timestep, chronic_scenario=chronic_scenario)
    if(modified_thermal_Limit):
        env._thermal_limit_a[ltc]=modified_thermal_Limit
    observation_space = env.observation_space
    sim = Grid2opSimulation(obs, action_space,observation_space, param_options=config["DEFAULT"], debug=False,
                            ltc=[ltc],simu_step=timestep)
                            #, plot=True, plot_folder="./alphaDeesp/tests/output_tempo")

    return sim,env


def test_save_red_dataframe():
    """Simple test, where a pandas.DataFrame gets saved, read and compared"""

    df = pd.DataFrame([[10.123123, -12.123123], [20.0123210, 91.1111]])
    df.to_csv("./test_read.csv")

    path_file = Path.cwd() / "test_read.csv"
    saved_df = pd.read_csv(path_file, index_col=0)

    print("They two dataframes are equal: ", are_dataframes_equal(df, saved_df))


def test_round_random_tests():
    arr = [12.312312, 11.094329]

    new_arr = [round(elem, 2) for elem in arr]
    print(new_arr)
    assert (new_arr == [12.31, 11.09])


def are_dataframes_equal(df1, df2):
    """Function that compares row by row, then field after field, 2 dataframes.
    If each value is identical, then they are equal

    Returns True if identical, False otherwise
    """

    generated_bag = []
    saved_bag = []

    for i, row in df1.iterrows():
        generated_bag.append(list(row))

    for i, row in df2.iterrows():
        saved_bag.append(list(row))

    # to properly compare the results, we round all floats to .XX 2 decimals
    generated_bag_rounded = []
    saved_bag_rounded = []

    for row in generated_bag:
        row_tab = []

        for elem in row:
            # print("elem = ", elem)
            # print("type = ", type(elem))
            if isinstance(elem, float):
                elem = round(elem, 2)

            if type(elem).__module__ == np.__name__:  # Has been identified as numpy array : has to be converted in list
                elem = elem.tolist()
            if type(elem) != list:
                if pd.isnull(elem): # NaN cant be equal to anything, even NaN themselves
                    elem = str(elem)
            row_tab.append(elem)

        generated_bag_rounded.append(row_tab)

    print("SEPARATOR ============================================================================================")

    for row in saved_bag:
        row_tab = []

        for elem in row:
            if isinstance(elem, float):
                elem = round(elem, 2)
            elif isinstance(elem, str):
                # string evaluation are used for arrays in string form, they get transformed into arrays
                elem = ast.literal_eval(elem)

            if type(elem).__module__ == np.__name__:  # Has been identified as numpy array : has to be converted in list
                elem = elem.tolist()
            if type(elem) != list:
                if pd.isnull(elem): # NaN cant be equal to anything, even NaN themselves
                    elem = str(elem)

            row_tab.append(elem)


        saved_bag_rounded.append(row_tab)

    print(generated_bag_rounded)
    print(saved_bag_rounded)

    return generated_bag_rounded == saved_bag_rounded


def test_integration_dataframe_results_with_line_9_cut():
    """
    In the initial state of the network, all substations are on busbar1
    Line 9 is between Node 4 and 5 [internal node ID indexing]
    Test
    """
    # import os
    # os.chdir('../../../')

    ltc = 9
    param_folder = "./alphaDeesp/tests/resources_for_tests_grid2op/l2rpn_2019_ltc_9"

    sim,env = build_sim(ltc, param_folder)
    df_of_g = sim.get_dataframe()
    g_over = OverFlowGraph(sim.topo, [ltc], df_of_g).get_graph()#sim.build_graph_from_data_frame([ltc])
    simulator_data = {"substations_elements": sim.get_substation_elements(),
                      "substation_to_node_mapping": sim.get_substation_to_node_mapping(),
                      "internal_to_external_mapping": sim.get_internal_to_external_mapping()}
    # create AlphaDeesp
    alphadeesp = AlphaDeesp(g_over, df_of_g, simulator_data=simulator_data)
    ranked_combinations = alphadeesp.get_ranked_combinations()
    expert_system_results, actions = sim.compute_new_network_changes(ranked_combinations)

    #expert_system_results.to_csv("alphaDeesp/tests/resources_for_tests_grid2op/END_RESULT_DATAFRAME_G2OP_LTC9_9CAPA_230_generated.csv")

    path_to_saved_end_result_dataframe = \
        Path.cwd() / "alphaDeesp/tests/resources_for_tests_grid2op/END_RESULT_DATAFRAME_G2OP_LTC9_9CAPA_230.csv"

    #expert_system_results.to_csv(path_to_saved_end_result_dataframe)
    saved_df = pd.read_csv(path_to_saved_end_result_dataframe, index_col=0)

    # List understandable format
    saved_df["Internal Topology applied "] = saved_df["Internal Topology applied "].str.replace(" ", ",")

    #print("The two dataframes are equal: ", are_dataframes_equal(expert_system_results, saved_df))
    assert are_dataframes_equal(expert_system_results[["Substation ID","Topology applied","Worsened line","Topology simulated score"]],
                                saved_df[["Substation ID","Topology applied","Worsened line","Topology simulated score"]])



def test_integration_dataframe_results_with_line_8_cut():
    """
    In the initial state of the network, all substations are on busbar1
    Line 8 is between Node 4 and 5 [internal node ID indexing]
    Test
    """

    # import os
    # os.chdir('../../../')

    ltc = 8
    param_folder = "./alphaDeesp/tests/resources_for_tests_grid2op/l2rpn_2019_ltc_8"

    sim,env = build_sim(ltc, param_folder)
    df_of_g = sim.get_dataframe()
    g_over = OverFlowGraph(sim.topo, [ltc], df_of_g).get_graph()
    simulator_data = {"substations_elements": sim.get_substation_elements(),
                      "substation_to_node_mapping": sim.get_substation_to_node_mapping(),
                      "internal_to_external_mapping": sim.get_internal_to_external_mapping()}
    # create AlphaDeesp
    alphadeesp = AlphaDeesp(g_over, df_of_g, simulator_data=simulator_data)
    ranked_combinations = alphadeesp.get_ranked_combinations()
    expert_system_results, actions = sim.compute_new_network_changes(ranked_combinations)

    path_to_saved_end_result_dataframe = \
        Path.cwd() / "alphaDeesp/tests/resources_for_tests_grid2op/END_RESULT_DATAFRAME_G2OP_LTC8_8CAPA_88.csv"

    # expert_system_results.to_csv(path_to_saved_end_result_dataframe)
    saved_df = pd.read_csv(path_to_saved_end_result_dataframe, index_col=0)

    # List understandable format
    saved_df["Internal Topology applied "] = saved_df["Internal Topology applied "].str.replace(" ", ",")

    #print("The two dataframes are equal: ", are_dataframes_equal(expert_system_results, saved_df))
    df1=expert_system_results[["Substation ID","Topology applied","Worsened line","Topology simulated score"]]
    df2=saved_df[["Substation ID","Topology applied","Worsened line","Topology simulated score"]]
    assert are_dataframes_equal(df1,df2)



def test_integration_dataframe_results_with_modified_substation4():
    """
    In the initial state of the network, all substations are on busbar1
    Line 9 is between Node 4 and 5 [internal node ID indexing]
    Test
    """

    # import os
    # os.chdir('../../../')

    timestep = 5
    ltc = 8
    param_folder = "./alphaDeesp/tests/resources_for_tests_grid2op/l2rpn_2019_ltc_8_modify_substation_4"
    config = configparser.ConfigParser()
    config.read("./alphaDeesp/tests/resources_for_tests_grid2op/config_for_tests.ini")

    ## Read Grid2op environment at timestep
    loader = Grid2opObservationLoader(param_folder)
    env, obs, action_space = loader.get_observation(timestep=timestep)

    ## Modify buses
    action = action_space({"set_bus": {'lines_ex_id': [(1, 2)], "lines_or_id": [(9, 2)]}})
    new_obs, reward, done, info = env.step(action)

    ## Build simulator and generate objects for alphadeesp
    sim = Grid2opSimulation(new_obs, action_space,env.observation_space, param_options=config["DEFAULT"], debug=False,
                            ltc=[ltc],simu_step=timestep)
    df_of_g = sim.get_dataframe()
    g_over = OverFlowGraph(sim.topo, [ltc], df_of_g).get_graph()
    simulator_data = {"substations_elements": sim.get_substation_elements(),
                      "substation_to_node_mapping": sim.get_substation_to_node_mapping(),
                      "internal_to_external_mapping": sim.get_internal_to_external_mapping()}

    ## Launch AlphaDeesp and get expert results
    alphadeesp = AlphaDeesp(g_over, df_of_g, simulator_data=simulator_data)
    ranked_combinations = alphadeesp.get_ranked_combinations()
    expert_system_results, actions = sim.compute_new_network_changes(ranked_combinations)
    # =============

    # Read desired results
    path_to_saved_end_result_dataframe = \
        Path.cwd() / "alphaDeesp/tests/resources_for_tests_grid2op/END_RESULT_DATAFRAME_G2OP_MODIFIED_SUBSTATION4.csv"
    #expert_system_results.to_csv(path_to_saved_end_result_dataframe)
    saved_df = pd.read_csv(path_to_saved_end_result_dataframe, index_col=0)

    ## Properly compare the two dataframes
    # List understandable format
    saved_df["Internal Topology applied "] = saved_df["Internal Topology applied "].str.replace(" ", ",")
    #print("The two dataframes are equal: ", are_dataframes_equal(expert_system_results, saved_df))
    assert are_dataframes_equal(expert_system_results[["Substation ID","Topology applied","Worsened line","Topology simulated score"]],
                                saved_df[["Substation ID","Topology applied","Worsened line","Topology simulated score"]])


def test_integration_dataframe_results_with_case_14_realistic():
    """
    In the initial state of the network, all substations are on busbar1
    Test
    """

    # import os
    # os.chdir('../../../')

    ltc = 4
    chronic_scenario = "000"
    timestep = 518
    param_folder = "./alphaDeesp/ressources/parameters/rte_case14_realistic" # We go directly in the folder to avoid double storing of "heavy" data
    config_file = "./alphaDeesp/tests/resources_for_tests_grid2op/config_for_tests.ini"

    sim,env = build_sim(ltc, param_folder, config_file = config_file, timestep=timestep, chronic_scenario=chronic_scenario)
    df_of_g = sim.get_dataframe()
    g_over = OverFlowGraph(sim.topo, [ltc], df_of_g).get_graph()

    simulator_data = {"substations_elements": sim.get_substation_elements(),
                      "substation_to_node_mapping": sim.get_substation_to_node_mapping(),
                      "internal_to_external_mapping": sim.get_internal_to_external_mapping()}
    # create AlphaDeesp
    alphadeesp = AlphaDeesp(g_over, df_of_g, simulator_data=simulator_data)
    ranked_combinations = alphadeesp.get_ranked_combinations()
    expert_system_results, actions = sim.compute_new_network_changes(ranked_combinations)

    # expert_system_results.to_csv("alphaDeesp/tests/resources_for_tests_grid2op/END_RESULT_DATAFRAME_G2OP_CASE14_REALISTIC_generated.csv")

    path_to_saved_end_result_dataframe = \
        Path.cwd() / "alphaDeesp/tests/resources_for_tests_grid2op/END_RESULT_DATAFRAME_G2OP_CASE14_REALISTIC.csv"

    saved_df = pd.read_csv(path_to_saved_end_result_dataframe, index_col=0)

    # List understandable format
    saved_df["Internal Topology applied "] = saved_df["Internal Topology applied "].str.replace(" ", ",")

    #print("The two dataframes are equal: ", are_dataframes_equal(expert_system_results, saved_df))
    assert are_dataframes_equal(expert_system_results[["Substation ID","Topology applied","Worsened line","Topology simulated score"]],
                                saved_df[["Substation ID","Topology applied","Worsened line","Topology simulated score"]])


def test_integration_dataframe_results_no_hubs():
    """
    In the initial state of the network, all substations are on busbar1
    No hubs are detected by alphadeesp - result should be an empty dataframe
    Test
    """

    # import os
    # os.chdir('../../../')

    ltc = 9
    chronic_scenario = "i"#None#
    timestep = 0#1
    param_folder = "./alphaDeesp/tests/resources_for_tests_grid2op/l2rpn_2019_nohubs"#"rte_case14_realistic"#
    config_file = "./alphaDeesp/tests/resources_for_tests_grid2op/config_for_tests.ini"

    sim,env = build_sim(ltc, param_folder, config_file = config_file, timestep=timestep, chronic_scenario=chronic_scenario)
    df_of_g = sim.get_dataframe()

    # emulate that edges 4->3, 5->10, 12->13 are grey edges, hence not to be considered in the graphs later.
    # hence no hubs will be identified
    df_of_g["delta_flows"].iloc[6] = 0.0
    df_of_g["gray_edges"].iloc[6] = True

    df_of_g["delta_flows"].iloc[12] = 0.0
    df_of_g["gray_edges"].iloc[12] = True

    df_of_g["delta_flows"].iloc[19] = 0.0
    df_of_g["gray_edges"].iloc[19] = True
    ####

    g_over = OverFlowGraph(sim.topo, [ltc], df_of_g).get_graph()

    simulator_data = {"substations_elements": sim.get_substation_elements(),
                      "substation_to_node_mapping": sim.get_substation_to_node_mapping(),
                      "internal_to_external_mapping": sim.get_internal_to_external_mapping()}
    # create AlphaDeesp
    alphadeesp = AlphaDeesp(g_over, df_of_g, simulator_data=simulator_data)
    ranked_combinations = alphadeesp.get_ranked_combinations()
    expert_system_results, actions = sim.compute_new_network_changes(ranked_combinations)

    # expert_system_results.to_csv("alphaDeesp/tests/resources_for_tests_grid2op/END_RESULT_DATAFRAME_G2OP_NO_HUBS_generated.csv")
    # expert_system_results.to_csv(r'D:\RTE\ExpertOp4Grid\5 - Résultats\2021_Timestep0_errors\generated_df.csv')

    path_to_saved_end_result_dataframe = \
        Path.cwd() / "alphaDeesp/tests/resources_for_tests_grid2op/END_RESULT_DATAFRAME_G2OP_NO_HUBS.csv"

    saved_df = pd.read_csv(path_to_saved_end_result_dataframe, index_col=0)

    # List understandable format
    saved_df["Internal Topology applied "] = saved_df["Internal Topology applied "].str.replace(" ", ",")

    #print("The two dataframes are equal: ", are_dataframes_equal(expert_system_results, saved_df))
    assert are_dataframes_equal(expert_system_results[["Substation ID","Topology applied","Worsened line","Topology simulated score"]],
                                saved_df[["Substation ID","Topology applied","Worsened line","Topology simulated score"]])


def test_integration_l2rpn_wcci_2020_computation_time():
    """
    In the initial state of the network, all substations are on busbar1
    Test
    """

    # import os
    # os.chdir('../../../')

    # Time threshold
    max_elapsed_time = 60 # seconds

    # Configuration
    ltc = 13
    chronic_scenario = "Scenario_february_069"
    timestep = 100
    param_folder = "./alphaDeesp/tests/resources_for_tests_grid2op/l2rpn_wcci_2020"
    config_file = "./alphaDeesp/tests/resources_for_tests_grid2op/config_for_tests.ini"
    sim,env = build_sim(ltc, param_folder, config_file = config_file, timestep=timestep, chronic_scenario=chronic_scenario)

    # Starting time
    start = time.time()

    # Simulation objects
    df_of_g = sim.get_dataframe()
    g_over = OverFlowGraph(sim.topo, [ltc], df_of_g).get_graph()
    simulator_data = {"substations_elements": sim.get_substation_elements(),
                      "substation_to_node_mapping": sim.get_substation_to_node_mapping(),
                      "internal_to_external_mapping": sim.get_internal_to_external_mapping()}

    # create AlphaDeesp
    printer = None
    custom_layout = sim.get_layout()
    alphadeesp = AlphaDeesp(g_over, df_of_g, simulator_data, sim.substation_in_cooldown, debug=False)

    # End time
    elapsed_time = time.time() - start
    print("the computation time is: " + str(elapsed_time))
    assert (elapsed_time <= max_elapsed_time)

def test_double_lines_wcci_2020():
    """
    In the initial state of the network, all substations are on busbar1
    Test
    """

    # import os
    # os.chdir('../../../')

    # Time threshold
    max_elapsed_time = 60 # seconds

    # Configuration
    ltc = 27
    chronic_scenario = "Scenario_february_069"
    timestep = 100
    param_folder = "./alphaDeesp/tests/resources_for_tests_grid2op/l2rpn_wcci_2020"
    config_file = "./alphaDeesp/tests/resources_for_tests_grid2op/config_for_tests.ini"

    modified_thermal_Limit=150#the flow in the line is about 158amps
    sim,env = build_sim(ltc, param_folder, config_file = config_file, timestep=timestep, chronic_scenario=chronic_scenario,modified_thermal_Limit=modified_thermal_Limit)

    # Starting time
    start = time.time()

    # Simulation objects
    df_of_g = sim.get_dataframe()
    g_over = OverFlowGraph(sim.topo, [ltc], df_of_g).get_graph()

    if(len(df_of_g)!=g_over.number_of_edges()):
        print("some edges were not properly added to the graph")
        assert(len(df_of_g)==g_over.number_of_edges())
    simulator_data = {"substations_elements": sim.get_substation_elements(),
                      "substation_to_node_mapping": sim.get_substation_to_node_mapping(),
                      "internal_to_external_mapping": sim.get_internal_to_external_mapping()}

    # create AlphaDeesp
    printer = None
    custom_layout = sim.get_layout()
    alphadeesp = AlphaDeesp(g_over, df_of_g, simulator_data, sim.substation_in_cooldown, debug=False)

    # End time

    print("AlphaDeesp succeeded for an overflow graph with double lines")




import setuptools
from setuptools import setup

pkgs = {
    "required": [
        "docutils>=0.16",
        "graphviz>=0.14",
        "Grid2Op>=1.6.4",
        "lightsim2grid>=0.5.5",
        "networkx>=2.4",
        "numpy>=1.18.4",
        "pandapower>=2.2.2",
        "scipy>=1.6.0",#needed for pandapower compatibility for now
        "pandas>=1.0.4",
        "pathlib>=1.0.1",
        "pydot>=1.4.1",
        #"pygame==1.9.6",
        "matplotlib>=3.3.3",
        "pytest>=4.4.2",
        "Sphinx>=3.1.2"
        #numba==0.49.1
    ],
    "extras": {
        "optional": [
            "oct2py>=5.0.4",
            "pypower>=5.1.4",
            "pypownet>=2.2.0"
        ]
    }
}

setup(name='ExpertOp4Grid',
      version='0.2.7',
      description='Expert analysis algorithm for solving overloads in a powergrid',
      long_description_content_type="text/markdown",
      classifiers=[
          'Development Status :: 4 - Beta',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
          "Intended Audience :: Developers",
          "Intended Audience :: Education",
          "Intended Audience :: Science/Research",
          "Natural Language :: English"
      ],
      keywords='ML powergrid optmization RL power-systems',
      author='Antoine Marot',
      author_email='antoine.marot@rte-france.com',
      url="https://github.com/marota/ExpertOp4Grid/",
      download_url = 'https://github.com/marota/ExpertOp4Grid/archive/refs/tags/0.1.3.post1.tar.gz',
      license='Mozilla Public License 2.0 (MPL 2.0)',
      packages=setuptools.find_packages(),
      extras_require=pkgs["extras"],
      include_package_data=True,
      install_requires=pkgs["required"],
      zip_safe=False,
      entry_points={'console_scripts': ['expertop4grid=alphaDeesp.main:main']},
)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of ExpertOp4Grid, an expert system approach to solve flow congestions in power grids

from pprint import pprint
import ast

import numpy as np
import pandas as pd
from math import fabs
import networkx as nx
from grid2op.dtypes import dt_int

from grid2op.PlotGrid import PlotMatplot

from alphaDeesp.core.simulation import Simulation
from alphaDeesp.core.network import Network
from alphaDeesp.core.elements import OriginLine, Consumption, Production, ExtremityLine
from alphaDeesp.core.printer import Printer
from alphaDeesp.core.graphsAndPaths import OverFlowGraph,PowerFlowGraph


class Grid2opSimulation(Simulation):
    def compute_layout(self):
        try:
            layout = self.param_options['CustomLayout']
            # Conversion from string to list
            layout = ast.literal_eval(layout)
            print("WARNING : A CustomLayout has been given in config.ini. This layout will be set for the simulator")
        except:
            try:
                # Grid2op Layout if exists
                layout = list(self.obs.grid_layout.values())
                print("WARNING : No CustomLayout has been given in config.ini. The grid_layout in Grid2Op structure will be set for the simulator")
            except:
                layout = [(-280, -81), (-100, -270), (366, -270), (366, -54), (-64, -54), (-64, 54), (366, 0),
                          (438, 0), (326, 54), (222, 108), (79, 162), (-152, 270), (-64, 270), (222, 216),
                          (-280, -151), (-100, -340), (366, -340), (390, -110), (-14, -104), (-184, 54), (400, -80),
                          (438, 100), (326, 140), (200, 8), (79, 12), (-152, 170), (-70, 200), (222, 200)]
                print("WARNING : No CustomLayout has been given in config.ini and no grid_layout has been found in Grid2op data. Default layout is set and might cause plotting errors : "+str(layout))
        return layout

    def get_layout(self):
        return self.layout

    def __init__(self, obs, action_space, observation_space, param_options=None, debug = False, ltc=[9],other_ltc=[], plot=False, plot_folder = None,reward_type=None,simu_step=0):
        super().__init__()

        # Get Grid2op objects
        if ltc is None:
            ltc = [9]
        self.plot_folder=None
        if plot: # Manual mode
            self.plot_folder = plot_folder
            self.printer = Printer(plot_folder)
        self.obs = obs
        self.lines_outaged = np.argwhere(self.obs.line_status == False)[:, 0]
        self.obs_linecut = None
        self.action_space = action_space#env.action_space
        self.observation_space = observation_space#env.observation_space
        #self.env=env
        self.no_overflow_disc = self.obs._obs_env.no_overflow_disconnection # Keep it in memory to activate and deactivate during computation steps

        # Get Alphadeesp configuration
        self.ltc = ltc
        self.other_ltc = other_ltc
        self.substation_in_cooldown=self.get_substation_in_cooldown()
        self.param_options = param_options
        self.reward_type=reward_type
        self.args_number_of_simulated_topos = param_options["totalnumberofsimulatedtopos"]
        self.args_inner_number_of_simulated_topos_per_node = param_options["numberofsimulatedtopospernode"]

        print("Number of generators of the powergrid: {}".format(self.obs.n_gen))
        print("Number of loads of the powergrid: {}".format(self.obs.n_load))
        print("Number of powerline of the powergrid: {}".format(self.obs.n_line))
        print("Number of elements connected to each substations in the powergrid: {}".format(self.obs.sub_info))
        print("Total number of elements: {}".format(self.obs.dim_topo))
        self.internal_to_external_mapping = {}
        self.external_to_internal_mapping = {}
        self.substations_elements = {}

        # Layout of the grid
        self.layout = self.compute_layout()

        # Compute data structure representing grid an dtopology
        self.topo = None
        self.topo_linecut = None
        self.df = None
        self.simu_step = simu_step
        self.load()
        self.save_bag = []

    def load(self):
        self.load_from_observation(self.obs, self.ltc+self.other_ltc)

    def load_from_observation(self, obs, linesToDisconnect):
        #self.obs = obs
        self.topo = self.extract_topo_from_obs(obs)
        self.topo_linecut = None
        self.df = self.create_df(self.topo, linesToDisconnect)
        self.internal_to_external_mapping = {}
        self.external_to_internal_mapping = {}
        self.substations_elements = {}
        self.create_and_fill_internal_structures(obs, self.df)

    def set_simu_step(self,simu_step):
        self.simu_step=simu_step

    def get_substation_elements(self):
        return self.substations_elements

    def get_substation_in_cooldown(self):
        return [i for i in range(self.obs.n_sub) if (self.obs.time_before_cooldown_sub[i]>=1)]

    def get_reference_topovec_sub(self,sub):
        nelements=self.obs.sub_info[sub]
        topovec=[0 for i in range(nelements)]
        return topovec

    def get_overload_disconnection_topovec_subor(self,l):
        sub_or=self.obs.line_or_to_subid[l]
        position_at_sub=self.obs.line_or_to_sub_pos[l]
        current_topo_vec=self.obs.state_of(substation_id=sub_or)['topo_vect']

        new_topo_vec=current_topo_vec
        new_topo_vec[position_at_sub]=-1#to get line disconnection
        return sub_or,new_topo_vec

    def get_substation_to_node_mapping(self):
        pass

    def get_internal_to_external_mapping(self):
        return self.internal_to_external_mapping

    @staticmethod
    def merge_two_dicts(x, y):
        z = x.copy()   # start with x's keys and values
        z.update(y)    # modifies z with y's keys and values & returns None
        return z

    def get_action_from_topo(self, substation_id, new_conf, obs):
        final_dict = {}
        i = 0
        objects = obs.get_obj_connect_to(substation_id=substation_id)
        for gen_id in objects['generators_id']:
            if "generators_id" not in final_dict:
                final_dict["generators_id"] = []
            final_dict["generators_id"].append((gen_id, new_conf[i]))
            i += 1
        for load_id in objects['loads_id']:
            if "loads_id" not in final_dict:
                final_dict["loads_id"] = []
            final_dict["loads_id"].append((load_id, new_conf[i]))
            i += 1
        for line_id in objects['lines_or_id']:
            if "lines_or_id" not in final_dict:
                final_dict["lines_or_id"] = []
            final_dict["lines_or_id"].append((line_id, new_conf[i]))
            i += 1
        for line_id in objects['lines_ex_id']:
            if "lines_ex_id" not in final_dict:
                final_dict["lines_ex_id"] = []
            final_dict["lines_ex_id"].append((line_id, new_conf[i]))
            i += 1
        # print(final_dict)
        return self.action_space({"set_bus": final_dict})

    def compute_new_network_changes(self, ranked_combinations):
        """
        This function takes a dataframe ranked_combinations,
        For each combination it computes a simulation step in Grid2op by following the given combinations
        Number of tested combinations and topo per node is given in alphadeesp parameters
        :returns pandas.DataFrame with results of simulations
        """
        print("\n##############################################################################")
        print("##########...........COMPUTE NEW NETWORK CHANGES..........####################")
        print("##############################################################################")
        end_result_dataframe = self.create_end_result_empty_dataframe()
        actions = []
        j = 0
        for df in ranked_combinations:
            ii = 0
            if j == int(self.args_number_of_simulated_topos):
                break
            for i, row in df.iterrows():
                if ii == int(self.args_inner_number_of_simulated_topos_per_node):
                    break
                obs = self.obs

                # realign timestep if still inconsistent alignment in Grid2op when chaining simulations
                time_step_simu = max(self.simu_step-obs.current_step,0)

                # target_node = row["node"] + 1
                internal_target_node = row["node"]
                # target_node = row["node"]
                alphaDeesp_Internal_topo=np.array([n for n in row["topology"]])

                new_conf = np.array([n + 1 for n in row["topology"]])
                if(len(alphaDeesp_Internal_topo)==1):#this is a line to disconnect, not a topology to change
                    l=alphaDeesp_Internal_topo[0]
                    #sub_id,new_conf_grid2op=self.get_overload_disconnection_topovec_subor(l)
                    new_conf=[l]
                    #action = self.action_space({"set_bus": {"substations_id": [(sub_id, new_conf_grid2op)] }})


                score_topo = i
                print("###########"" Compute new network changes on node [{}] with new topo [{}] ###########"
                      .format(internal_target_node, new_conf))


                if(len(alphaDeesp_Internal_topo)==1):#this is a line to disconnect, not a topology to change
                    new_conf_grid2op=[l]
                    new_conf=[l]
                    action = self.action_space({"set_line_status": [(l, -1)]})
                else:
                    action = self.get_action_from_topo(internal_target_node, new_conf, obs)
                    new_conf_grid2op = list(action.effect_on(substation_id=internal_target_node)[
                                                'set_bus'])  # grid2op conf is different from alphadeesp conf, because the elements are ordered differently
                actions.append(action)


                # simulation should be done on same timestep than observation, but with time alignment correction
                virtual_obs, reward, done, info = self.obs.simulate(action, time_step = time_step_simu)

                score_data=self.compute_one_network_change_score_data(obs,virtual_obs,done,info,new_conf,internal_target_node,alphaDeesp_Internal_topo,new_conf_grid2op,score_topo)
                #print(score_data)
                #end_result_dataframe.loc[max_index] = score_data
                #max_index = end_result_dataframe.shape[0]  # rows
                #end_result_dataframe=end_result_dataframe.append(pd.Series(score_data, index=end_result_dataframe.columns ),ignore_index=True)
                end_result_dataframe.loc[len(end_result_dataframe)]=pd.Series(score_data, index=end_result_dataframe.columns )
                ii += 1
                j += 1

        end_result_dataframe.to_csv("./END_RESULT_DATAFRAME.csv", index=True)

        # Case there are no hubs --> action do nothing
        if len(actions) == 0:
            actions = [self.action_space()]
        return end_result_dataframe, actions

    def compute_one_network_change_score_data(self, obs,virtual_obs,done,info,new_conf,internal_target_node,alphaDeesp_Internal_topo,new_conf_grid2op,score_topo):
        # Same as in Pypownet, this is not what we would want though, as we do the work for only one ltc
        only_line = self.ltc[0]
        line_state_before = obs.state_of(line_id=only_line)
        line_state_after = virtual_obs.state_of(line_id=only_line)
        flow_before = line_state_before["origin"]["p"]
        flow_after = line_state_after["origin"]["p"]
        delta_flow = flow_before - flow_after

        if done:  # Game over: no need to compute further operations
            worsened_line_ids = []
            simulated_score = 0
            redistribution_prod = float('nan')
            redistribution_load = float('nan')
            efficacity = float('nan')

        else:
            # Fill save bag with observations for further analysis (detailed graph)
            name = "".join(str(e) for e in alphaDeesp_Internal_topo)#"".join(str(e) for e in new_conf)
            name = str(internal_target_node) + "_" + name
            self.save_bag.append([name, virtual_obs])
            worsened_line_ids = self.create_boolean_array_of_worsened_line_ids(obs, virtual_obs,
                                                                               self.obs._obs_env.parameters.NB_TIMESTEP_COOLDOWN_LINE)
            simulated_score = score_changes_between_two_observations(self.ltc, obs, virtual_obs,
                                                                     self.obs._obs_env.parameters.NB_TIMESTEP_COOLDOWN_LINE)

            # update simulated score to 0 in case
            # our line got disconnected, starting a cascading failure
            if (bool(info['disc_lines'][self.ltc]>=0)):#other line disconnections are already accounted in worsened lines
            #if (info['disc_lines'].any()):
                simulated_score = 0

            redistribution_prod = np.sum(np.absolute(virtual_obs.prod_p - obs.prod_p))

            TotalProd = np.nansum(virtual_obs.prod_p)
            Losses = np.nansum(np.abs(virtual_obs.p_or + virtual_obs.p_ex))
            ExpectedNewLoad = TotalProd - Losses
            redistribution_load = (np.sum(virtual_obs.load_p) - ExpectedNewLoad)  # / np.sum(old_obs.load_p)

            if simulated_score in [4, 3, 2, 1]:  # success
                efficacity = fabs(delta_flow / virtual_obs.rho[self.ltc[0]])
                if (self.reward_type is not None) and (self.reward_type in info["rewards"]):  # & (simulated_score==4):
                    # dans le cas ou on resoud bien les contraintes, on prend la reward L2RPN
                    efficacity = info["rewards"][self.reward_type]
            else:  # failure
                efficacity = -fabs(delta_flow / virtual_obs.rho[self.ltc[0]])

        # To store in data frame
        score_data = [only_line,
                      flow_before,
                      flow_after,
                      delta_flow,
                      worsened_line_ids,
                      redistribution_prod,
                      redistribution_load,
                      alphaDeesp_Internal_topo,  # alphaDeesp internal topology format
                      new_conf_grid2op,
                      # new_conf,#we prefer to have the backend conf definition, rather than alphadeesp one
                      internal_target_node,
                      1,  # category hubs?
                      score_topo,
                      simulated_score,
                      efficacity]
        # print(score_data)
        return score_data

    @staticmethod
    def create_boolean_array_of_worsened_line_ids(old_obs, new_obs,nb_timestep_cooldown_line_param):
        res = []
        n_lines=len(new_obs.rho)

        #for old, new in zip(old_obs, new_obs):
        for l in range(n_lines):
            if new_obs.rho[l] > 1 and old_obs.rho[l] > 1 and new_obs.rho[l] > 1.05 * old_obs.rho[l]:
                res.append(1)
            elif new_obs.rho[l] > 1 > old_obs.rho[l]:
                res.append(1)
            elif ((new_obs.time_before_cooldown_line[l] - old_obs.time_before_cooldown_line[l])>nb_timestep_cooldown_line_param): #line got into cascading failure
                res.append(1)
            else:
                res.append(0)
        res = np.array(res)
        res = list(np.where(res == 1))
        res = res[0]
        if res.size == 0:
            res = []
        elif isinstance(res, np.ndarray):
            res = list(res)
        return res

    def create_and_fill_internal_structures(self, obs, df):
        """This function fills multiple structures:
        self.substation_elements, self.substation_to_node_mapping, self.internal_to_external_mapping
        @:arg observation, df"""
        # ################ PART I : fill self.internal_to_external_mapping
        substations_list = list(obs.name_sub)
        # we create mapping from external representation to internal.
        for i, substation_id in enumerate(substations_list):
            self.internal_to_external_mapping[i] = substation_id
        if self.internal_to_external_mapping:
            self.external_to_internal_mapping = self.invert_dict_keys_values(self.internal_to_external_mapping)

        # ################ PART II : fill self.substation_elements
        for substation_id in self.internal_to_external_mapping.keys():
            elements_array = []
            objects = obs.get_obj_connect_to(substation_id=substation_id)
            for gen_id in objects['generators_id']:
                gen_state = obs.state_of(gen_id=gen_id)
                elements_array.append(Production(gen_state['bus']-1, gen_state['p']))
            for load_id in objects['loads_id']:
                load_state = obs.state_of(load_id=load_id)
                elements_array.append(Consumption(load_state['bus']-1, load_state['p']))
            for line_id in objects['lines_or_id']:
                line_state = obs.state_of(line_id=line_id)
                orig = line_state['origin']
                ext = line_state['extremity']
                dest = ext['sub_id']
                elements_array.append(self.get_model_obj_from_or(self.df, substation_id, dest, orig['bus']-1))
            for line_id in objects['lines_ex_id']:
                line_state = obs.state_of(line_id=line_id)
                orig = line_state['origin']
                ext = line_state['extremity']
                dest = orig['sub_id']
                elements_array.append(self.get_model_obj_from_ext(self.df, substation_id, dest, ext['bus']-1))
            self.substations_elements[substation_id] = elements_array
        # pprint(self.substations_elements)

    @staticmethod
    def extract_topo_from_obs(obs):
        """This function, takes an obs an returns a dict with all topology information"""
        d = {
            "edges": {},
            "nodes": {}
        }
        nsub = obs.n_sub
        nodes_ids = list(range(nsub))
        idx_or = obs.line_or_to_subid
        idx_ex = obs.line_ex_to_subid
        prods_ids = obs.gen_to_subid
        loads_ids = obs.load_to_subid
        are_prods = [node_id in prods_ids for node_id in nodes_ids]
        are_loads = [node_id in loads_ids for node_id in nodes_ids]
        current_flows = obs.p_or  # Flow at the origin of power line is taken

        # Repartition of prod and load in substations
        prods_values = obs.prod_p
        loads_values = obs.load_p
        gens_ordered_by_subid = np.argsort(obs.gen_to_subid)
        loads_ordered_by_subid = np.argsort(obs.load_to_subid)
        prods_values = prods_values[gens_ordered_by_subid]
        loads_values = loads_values[loads_ordered_by_subid]

        # Store topo in dictionary
        d["edges"]["idx_or"] = [x for x in idx_or]
        d["edges"]["idx_ex"] = [x for x in idx_ex]
        d["edges"]["init_flows"] = current_flows
        d["nodes"]["are_prods"] = are_prods
        d["nodes"]["are_loads"] = are_loads
        d["nodes"]["prods_values"] = prods_values
        d["nodes"]["loads_values"] = loads_values

        # Debug
        # for key in d.keys():
        #     print(key)
        #     for key2 in d[key].keys():
        #         print(key2)
        #         print(d[key][key2])
        return d

    def cut_lines_and_recomputes_flows(self, ids: list):
        """This functions cuts lines: [ids], simulates and returns new line flows"""

        # First, set parameter to avoid disconnection
        self.obs._obs_env._parameters.NO_OVERFLOW_DISCONNECTION = True
        self.obs._obs_env._no_overflow_disconnection=True
        max_line_actions=self.obs._obs_env.parameters.MAX_LINE_STATUS_CHANGED

        self.obs._obs_env._parameters.MAX_LINE_STATUS_CHANGED = 999

        # Set action which disconects the specified lines (by ids)
        deconexion_action = self.action_space({"set_line_status": [(id_, -1) for id_ in ids]})

        #realign timestep if still inconsistent alignment in Grid2op when chaining simulations
        time_step_simu = max(self.simu_step-self.obs.current_step,0)

        #simulation should be done on same timestep than observation, but with time alignment correction
        obs_linecut, reward, done, info = self.obs.simulate(deconexion_action, time_step = time_step_simu) #, time_step = 0)
        # Storage of new observation to access features in other function
        self.obs_linecut = obs_linecut
        self.topo_linecut = self.extract_topo_from_obs(self.obs_linecut)
        self.lines_outaged_cut = np.argwhere(self.obs_linecut.line_status == False)[:, 0]

        # Get new flow simulated
        new_flow = self.obs_linecut.p_or

        # Graph building
        # self.g_pow_prime = self.build_powerflow_graph(self.obs_cutted)

        # Finaly, reset previous parameter
        self.obs._obs_env._no_overflow_disconnection = self.no_overflow_disc
        self.obs._obs_env._parameters.NO_OVERFLOW_DISCONNECTION = self.no_overflow_disc
        self.obs._obs_env._parameters.MAX_LINE_STATUS_CHANGED=max_line_actions

        return new_flow

    def build_powerflow_graph_beforecut(self):
        """
        Builds a graph of the grid and its powerflow before the lines are cut
        :return: NetworkX Graph of representing the grid
        """
        lines_cut = np.argwhere(self.obs.line_status == False)[:, 0]
        #g = build_powerflow_graph(self.topo, lines_cut)
        return PowerFlowGraph(self.topo, lines_cut).g

    def build_powerflow_graph_aftercut(self):
        """
        Builds a graph of the grid and its powerflow after the lines have been cut
        :return: NetworkX Graph of representing the grid
        """
        lines_cut = np.argwhere(self.obs_linecut.line_status == False)[:, 0]
        #g = build_powerflow_graph(self.topo_linecut, lines_cut)
        return PowerFlowGraph(self.topo_linecut, lines_cut).g

    def get_dataframe(self):
        """
        :return: pandas dataframe with topology information before and after line cutting
        """
        return self.df

    def isAntenna(self):
        linesAtBusbar_dic = self.getLinesAtSubAndBusbar()

        for sub in linesAtBusbar_dic.keys():
            linesAtBusbar = linesAtBusbar_dic[sub]
            if len(linesAtBusbar) <= 1:
                return sub  # this is an Antenna
        return None

    def isDoubleLine(self):
        ltc = self.ltc[0]
        obs = self.obs

        sub_or = int(obs.line_or_to_subid[ltc])
        sub_ex = int(obs.line_ex_to_subid[ltc])

        linesOr_atSubOr = list(obs.get_obj_connect_to(substation_id=sub_or)['lines_or_id'])
        linesEx_atSubOr = list(obs.get_obj_connect_to(substation_id=sub_or)['lines_ex_id'])

        lines_atSubOr=linesOr_atSubOr+linesEx_atSubOr

        linesOr_atSubEx = list(obs.get_obj_connect_to(substation_id=sub_ex)['lines_or_id'])
        linesEx_atSubEx = list(obs.get_obj_connect_to(substation_id=sub_ex)['lines_ex_id'])

        lines_atSubEx = linesOr_atSubEx+linesEx_atSubEx

        Common_lines=[l for l in lines_atSubEx if (l in lines_atSubOr) and (l != ltc)]

        if(len(Common_lines)==0):
            return None
        else:
            return Common_lines


    def getLinesAtSubAndBusbar(self):
        ltc = self.ltc[0]
        obs=self.obs
        linesAtBusbar_dic = {}

        sub_or = int(obs.line_or_to_subid[ltc])
        sub_ex = int(obs.line_ex_to_subid[ltc])

        busBarOr = obs.state_of(line_id=ltc)['origin']['bus']
        busBarEx = obs.state_of(line_id=ltc)['extremity']['bus']

        # we should check if another line is connected with our line of interest. Otherwise it is an antenna
        # For SubOr
        linesOr_atSubOr = obs.get_obj_connect_to(substation_id=sub_or)['lines_or_id']
        linesEx_atSubOr = obs.get_obj_connect_to(substation_id=sub_or)['lines_ex_id']

        linesAtBusbarOr = [l for l in linesOr_atSubOr if obs.state_of(line_id=l)['origin']['bus'] == busBarOr]
        linesAtBusbarOr += [l for l in linesEx_atSubOr if obs.state_of(line_id=l)['extremity']['bus'] == busBarOr]

        # For SubEx
        linesOr_atSubEx = obs.get_obj_connect_to(substation_id=sub_ex)['lines_or_id']
        linesEx_atSubEx = obs.get_obj_connect_to(substation_id=sub_ex)['lines_ex_id']

        linesAtBusbarEx = [l for l in linesOr_atSubEx if obs.state_of(line_id=l)['origin']['bus'] == busBarEx]
        linesAtBusbarEx += [l for l in linesEx_atSubEx if obs.state_of(line_id=l)['extremity']['bus'] == busBarEx]

        linesAtBusbar_dic[sub_or] = linesAtBusbarOr
        linesAtBusbar_dic[sub_ex] = linesAtBusbarEx
        return linesAtBusbar_dic

    def build_detailed_graph_from_internal_structure(self, lines_to_cut):
        """This function create a detailed graph from internal self structures as self.substations_elements..."""
        g = nx.MultiDiGraph()

        # Reduce busbar_ids by -1 (Grid2op: 1,2 / Pypownet: 0,1)
        # for key in self.substations_elements.keys():
        #     for elt in self.substations_elements[key]:
        #         elt.busbar_id -= 1

        network = Network(self.substations_elements)

        # print("Network = ", network)
        build_nodes_v2(g, network.nodes_prod_values)
        build_edges_v2(g, network.substation_id_busbar_id_node_id_mapping, self.substations_elements)
        # print("This graph is weakly connected : ", nx.is_weakly_connected(g))
        return g

    #def plot_grid_beforecut(self):
    #    """
    #    Plots the grid with Grid2op PlotHelper for Observations, before lines are cut
    #    :return: Figure
    #    """
    #    return self.plot_grid(self.obs, name = "g_pow")
#
    #def plot_grid_aftercut(self):
    #    """
    #    Plots the grid with Grid2op PlotHelper for Observations, after lines have been cut
    #    :return: Figure
    #    """
    #    return self.plot_grid(self.obs_linecut, name = "g_pow_prime")
#
    #def plot_grid_delta(self):
    #    """
    #    Plots the grid with alphadeesp.printer API for delta between Observations before and after lines have been cut
    #    :return: Figure
    #    """
    #    return self.plot_grid(None, name="g_overflow_print")
#
    #def plot_grid_from_obs(self, obs, name):
    #    """
    #    Plots the grid with Grid2op PlotHelper from given observation
    #    :return: Figure
    #    """
    #    return self.plot_grid(obs, name=name)

    def plot(self,obs,save_file_path):
        plot_helper = PlotMatplot(self.observation_space)
        fig_obs = plot_helper.plot_obs(obs, line_info='p')
        fig_obs.savefig(save_file_path)

    #def plot_grid(self, obs, name):
    #    type_ = "results"
    #    if name in ["g_pow", "g_overflow_print", "g_pow_prime"]:
    #        type_ = "base"
#
    #    if name == "g_overflow_print":  # Use printer API to plot g_over (graphviz/neato)
    #        overflowGraph = OverFlowGraph(self.topo, self.ltc, self.df)
    #        g_over = overflowGraph.g#self.build_graph_from_data_frame(self.ltc)
    #        self.printer.display_geo(g_over, self.get_layout(), name=name)
    #    else:   # Use grid2op plot functionalities to plot all other graphs
    #        output_name = self.printer.create_namefile("geo", name = name, type = type_)
    #        fig_obs = self.plot_helper.plot_obs(obs, line_info='p')
    #        fig_obs.savefig(output_name[1])
#
    def change_nodes_configurations(self, new_configurations, node_ids, env):
        change = []
        for (conf, node) in zip(new_configurations, node_ids):
            change.append((node, conf))
        action = self.action_space({"set_bus": {"substations_id": change}})
        new_obs, reward, done, info = env.step(action)
        self.obs = new_obs
        self.load()
        return new_obs



def build_nodes_v2(g, nodes_prod_values: list):
    """nodes_prod_values is a list of tuples, (graphical_node_id, prod_cons_total_value)
        prod_cons_total_value is a float.
        If the value is positive then it is a Production, if negative it is a Consumption
    """
    # print("IN FUNCTION BUILD NODES V2222222222", nodes_prod_values)
    for data in nodes_prod_values:
        # print("data = ", data)
        i = int(data[0])
        if data[1] is None or data[1] == "XXX":
            prod_minus_load = 0.0  # It will end up as a white node
        else:
            prod_minus_load = data[1]
        # print("prod_minus_load = ", prod_minus_load)
        if prod_minus_load > 0:  # PROD
            g.add_node(i, pin=True, prod_or_load="prod", value=str(prod_minus_load), style="filled",
                       fillcolor="#f30000")  # red color
        elif prod_minus_load < 0:  # LOAD
            g.add_node(i, pin=True, prod_or_load="load", value=str(prod_minus_load), style="filled",
                       fillcolor="#478fd0")  # blue color
        else:  # WHITE COLOR
            g.add_node(i, pin=True, prod_or_load="load", value=str(prod_minus_load), style="filled",
                       fillcolor="#ffffed")  # white color
        i += 1


def build_edges_v2(g, substation_id_busbar_id_node_id_mapping, substations_elements):
    # print("\nWE ARE IN BUILD EDGES V2")
    substation_ids = sorted(list(substations_elements.keys()))
    # loops through each substation, and creates an edge from (
    for substation_id in substation_ids:
        # print("\nSUBSTATION ID = ", substation_id)
        for element in substations_elements[substation_id]:
            # print(element)
            origin = None
            extremity = None
            if isinstance(element, OriginLine):
                # origin = substation_id
                origin = int(substation_id_busbar_id_node_id_mapping[substation_id][element.busbar_id])
                extremity = int(element.end_substation_id)
                # check if extremity on busbar1, if it is,
                # check with the substation substation_id_busbar_id_node_id_mapping dic what "graphical" node it is
                # print("substations_elements[extremity] = ", substations_elements[extremity])
                for elem in substations_elements[extremity]:
                    # if this true, we are talking about correct edge
                    if isinstance(elem, ExtremityLine) and elem.flow_value == element.flow_value:
                        if elem.busbar == 1:
                            extremity = substation_id_busbar_id_node_id_mapping[extremity][1]
                reported_flow = element.flow_value
            elif origin is None or extremity is None:
                continue
            # in case we get on an element that is Production or Consumption
            else:
                continue
            # print("origin = ", origin)
            # print("extremity = ", extremity)
            # print("reported_flow = ", reported_flow)
            pen_width = fabs(reported_flow[0]) / 10.0
            if pen_width < 0.01:
                pen_width = 0.1
            # print(f"#################### Edge created : ({origin}, {extremity}), with flow = {reported_flow},"
            #       f" pen_width = {pen_width} >>>")
            if reported_flow[0] > 0:  # RED
                g.add_edge(origin, extremity, capacity=float(reported_flow[0]), label=reported_flow[0], color="coral",
                           penwidth="%.2f" % pen_width)
            else:  # BLUE
                g.add_edge(origin, extremity, capacity=float(reported_flow[0]), label=reported_flow[0], color="blue",
                           penwidth="%.2f" % pen_width)
            g.add_edge(origin, extremity, capacity=float(reported_flow[0]), label=reported_flow[0])

#TO DO: check is line is disconnected in new_obs

def score_changes_between_two_observations(ltc, old_obs, new_obs,nb_timestep_cooldown_line_param=0):
    """This function takes two observations and computes a score to quantify the change between old_obs and new_obs.
    @:return int between [0 and 4]
    4: if every overload disappeared
    3: if an overload disappeared without stressing the network
    2: if at least 30% of an overload was relieved
    1: if an overload was relieved but another appeared and got worse
    0: if no overloads were alleviated or if it resulted in some load shedding or production distribution.
    """
    old_number_of_overloads = 0
    new_number_of_overloads = 0
    boolean_overload_worsened = []
    boolean_constraint_30percent_relieved = []
    boolean_constraint_relieved = []
    boolean_overload_created = []
    boolean_line_cascading_disconnection = ((new_obs.time_before_cooldown_line-old_obs.time_before_cooldown_line)>nb_timestep_cooldown_line_param)



    old_obs_lines_capacity_usage = old_obs.rho
    new_obs_lines_capacity_usage = new_obs.rho
    # ################################### PREPROCESSING #####################################
    for elem in old_obs_lines_capacity_usage:
        if elem > 1.0:
            old_number_of_overloads += 1

    for elem in new_obs_lines_capacity_usage:
        if elem > 1.0:
            new_number_of_overloads += 1

    # preprocessing for score 3 and 2
    line_id=0
    for old, new in zip(old_obs_lines_capacity_usage, new_obs_lines_capacity_usage):
        # preprocessing for score 3
        if (new > 1.05 * old) & (new > 1.0):  # if new > old > 1.0 it means it worsened an existing constraint
            boolean_overload_worsened.append(1)
        else:
            boolean_overload_worsened.append(0)

        # preprocessing for score 2
        if (old > 1.0) & (line_id in ltc) :  # if old was an overload:
            surcharge = old - 1.0
            diff = old - new
            percentage_relieved = diff * 100 / surcharge
            if percentage_relieved > 30.0:
                boolean_constraint_30percent_relieved.append(1)
            else:
                boolean_constraint_30percent_relieved.append(0)
        else:
            boolean_constraint_30percent_relieved.append(0)


        # preprocessing for score 1
        if (old > 1.0 > new) & (line_id in ltc):
            boolean_constraint_relieved.append(1)
        else:
            boolean_constraint_relieved.append(0)

        if old < 1.0 < new:
            boolean_overload_created.append(1)
        else:
            boolean_overload_created.append(0)

        line_id+=1

    boolean_overload_worsened = np.array(boolean_overload_worsened)
    boolean_constraint_30percent_relieved = np.array(boolean_constraint_30percent_relieved)
    boolean_constraint_relieved = np.array(boolean_constraint_relieved)
    boolean_overload_created = np.array(boolean_overload_created)

    redistribution_prod = np.sum(np.absolute(new_obs.prod_p - old_obs.prod_p))
    #redistribution_load = np.sum(np.absolute(new_obs.load_p - old_obs.load_p))#not exact in Grid2op if load are disconnected

    TotalProd=np.nansum(new_obs.prod_p)
    Losses=np.nansum(np.abs(new_obs.p_or+new_obs.p_ex))
    ExpectedNewLoad=TotalProd-Losses
    cut_load_percent=(np.sum(old_obs.load_p)-ExpectedNewLoad)/np.sum(old_obs.load_p)


    # ################################ END OF PREPROCESSING #################################
    # score 0 if no overloads were alleviated or if it resulted in some load shedding or production distribution.
    if old_number_of_overloads == 0:
        # print("return NaN: No overflow at initial state of grid")
        return float('nan')
    elif cut_load_percent > 0.01: # (boolean_overload_relieved == 0).all()
        # print("return 0: no overloads were alleviated or some load shedding occured.")
        return 0

    # score 1 if overload was relieved but another one appeared and got worse
    elif (boolean_constraint_relieved == 1).any() and ((boolean_overload_created == 1).any() or
                                                     (boolean_overload_worsened == 1).any() or (boolean_line_cascading_disconnection).any()):
        # print("return 1: our overload was relieved but another one appeared")
        return 1

    # 4: if every overload disappeared
    elif old_number_of_overloads > 0 and new_number_of_overloads == 0:
        # print("return 4: every overload disappeared")
        return 4

    # 3: if this overload disappeared without stressing the network, ie,
    # if an overload disappeared
    # and without worsening existing constraint
    # and no Loads that get cut
    elif (boolean_constraint_relieved == 1).any() and \
            (boolean_overload_worsened == 0).all():
        # and \ (new_obs.are_loads_cut == 0).all():
        # print("return 3: our overload disappeared without stressing the network")
        return 3

    # 2: if at least 30% of this overload was relieved
    elif (boolean_constraint_30percent_relieved == 1).any() and \
            (boolean_overload_worsened == 0).all():
        # print("return 2: at least 30% of our overload [{}] was relieved".format(
        #     np.where(boolean_overload_30percent_relieved == 1)[0]))
        return 2

    # score 0
    elif (boolean_constraint_30percent_relieved == 0).all() or \
            (boolean_overload_worsened == 1).any():
        return 0

    else:
        raise ValueError("Probleme with Scoring")



from math import fabs
import ast

import networkx as nx
import pypownet.environment
from pypownet.agent import *
from pypownet.environment import ElementType
import numpy as np

from alphaDeesp.core.elements import *
from alphaDeesp.core.network import Network
from alphaDeesp.core.simulation import Simulation
from alphaDeesp.core.printer import Printer


class PypownetSimulation(Simulation):
    def __init__(self, env, obs, action_space, param_options=None, debug=False, ltc=[9], plot_folder = None,isScoreFromBackend=False):
        super().__init__()
        print("PypownetSimulation object created...")

        if not param_options or param_options is None:
            raise AttributeError("\nparam_options are empty or None, meaning the config file is not properly read.")

        self.save_bag = []
        self.debug = debug
        self.args_number_of_simulated_topos = param_options["totalnumberofsimulatedtopos"]
        self.args_inner_number_of_simulated_topos_per_node = param_options["numberofsimulatedtopospernode"]
        self.grid = None
        self.df = None
        self.topo = None  # a dict create in retrieve topology
        self.ltc = ltc
        self.param_options = param_options
        self.printer = Printer(plot_folder)
        #############################
        self.environment = env
        self.action_space = action_space
        # Run one step in the environment
        self.obs = obs
        self.lines_outaged  = np.argwhere(self.obs.lines_status == 0)
        self.obs_linecut = None
        self.isScoreFromBackend=isScoreFromBackend

        # Layout of the grid
        self.layout = self.compute_layout()

        print("HARD OVERFLOW = ", self.environment.game.hard_overflow_coefficient)
        print("")

        observation_space = self.environment.observation_space


        #############################
        # new structures to omit querying Pypownet, they are filled in LOAD function.
        # for each substation, we get an array with (Prod, Cons, Line) Objects, representing the actual configuration
        self.substations_elements = {}
        self.substation_to_node_mapping = {}
        self.internal_to_external_mapping = {}  # d[internal_id] = external_name_id
        self.external_to_internal_mapping = {}  # d[external_id] = internal_name_id
        print("current chronic name = ", self.environment.game.get_current_chronic_name())
        print(self.obs)
        self.load()

    def compute_layout(self):
        try:
            layout = self.param_options['CustomLayout']
            # Conversion from string to list
            layout = ast.literal_eval(layout)
        except:
            layout = [(-280, -81), (-100, -270), (366, -270), (366, -54), (-64, -54), (-64, 54), (366, 0),
                    (438, 0), (326, 54), (222, 108), (79, 162), (-152, 270), (-64, 270), (222, 216),
                    (-280, -151), (-100, -340), (366, -340), (390, -110), (-14, -104), (-184, 54), (400, -80),
                    (438, 100), (326, 140), (200, 8), (79, 12), (-152, 170), (-70, 200), (222, 200)]
        return layout

    def get_layout(self):
        return self.layout

    def get_substation_elements(self):
        return self.substations_elements

    def get_substation_to_node_mapping(self):
        return self.substation_to_node_mapping

    def get_internal_to_external_mapping(self):
        return self.internal_to_external_mapping

    def get_dataframe(self):
        """
        :return: pandas dataframe with topology information before and after line cutting
        """
        return self.df

    def isAntenna(self):
        """TODO"""
        return None

    def isDoubleLine(self):
        """TODO"""
        return None

    def getLinesAtSubAndBusbar(self):
        """TODO"""
        return None

    def get_overload_disconnection_topovec_subor(self, l):
        """TODO"""
        return None,None

    def get_reference_topovec_sub(self,sub):
        """TODO"""
        return None

    def get_substation_in_cooldown(self):
        """TODO"""
        return None

    def compute_new_network_changes(self, ranked_combinations):
        """this function takes a dataframe ranked_combinations,
         for each combination it computes a simulation step in Pypownet with action:
         change nodes topo(combination)"""

        print("\n##############################################################################")
        print("##########...........COMPUTE NEW NETWORK CHANGES..........####################")
        print("##############################################################################")

        # the function score creates a Dataframe with sorted score for each topo change.
        # FINISHED
        end_result_dataframe = self.create_end_result_empty_dataframe()
        j = 0
        for df in ranked_combinations:
            ii = 0
            if j == int(self.args_number_of_simulated_topos):
                break
            for i, row in df.iterrows():
                if ii == int(self.args_inner_number_of_simulated_topos_per_node):
                    break
                saved_obs = self.obs
                # target_node = row["node"] + 1
                internal_target_node = row["node"]
                target_node = self.internal_to_external_mapping[row["node"]]
                # target_node = row["node"]
                new_conf = row["topology"]
                print("ROW KEYS = ", row.keys())
                score_topo = i
                print("#####################################################################################")
                print("###########"" Compute new network changes on node [{}] with new topo [{}] ###########"
                      .format(target_node, new_conf))
                print("#####################################################################################")
                current_conf, types = self.obs.get_nodes_of_substation(target_node)

                # target configuration represents the action to be taken to get from curr_conf to ==> new_conf.
                target_configuration = get_differencial_topology(new_conf, current_conf)

                # load this topo to pypownet
                action_space = self.environment.action_space
                observation_space = self.environment.observation_space
                # Create template of action with no switch activated (do-nothing action)
                action = action_space.get_do_nothing_action(as_class_Action=True)
                # This function fills the "action" with correct values.
                action_space.set_substation_switches_in_action(action=action, substation_id=target_node,
                                                               new_values=target_configuration)

                # action_space.set_lines_status_switch_from_id(action=action, line_id=[9], new_switch_value=1)
                # Run one step in the environment
                raw_obs,action, reward, *_ = self.environment.simulate(action)

                # if obs is None, error in the simulation of the next step
                if raw_obs is None:
                    print("Pypownet simulation returnt a None... Cannot process results...")
                    continue

                # transform raw_obs into Observation object. (Useful for prints, for debugging)
                obs = observation_space.array_to_observation(raw_obs)
                obs.get_lines_capacity_usage()
                if self.debug:
                    print(obs)

                print("old obs addr = ", hex(id(saved_obs)))
                print("new obs addr = ", hex(id(obs)))

                # this is used to display graphs at the end. Check main.
                name = "".join(str(e) for e in new_conf)
                name = str(internal_target_node) + "_" + name

                # for flow, origin, ext, in zip(obs.active_flows_origin, obs.lines_or_nodes, obs.lines_ex_nodes)

                self.save_bag.append([name, obs])

                delta_flow = saved_obs.active_flows_origin[self.ltc[0]] - \
                             obs.active_flows_origin[self.ltc[0]]

                print("self.lines to cut[0] = ", self.ltc[0])
                print("self.obs.status line = ", self.obs.lines_status)
                print(" simulated obs  = ", obs.lines_status)
                print("saved flow = ", list(saved_obs.active_flows_origin.astype(int)))
                print("current flow = ", list(obs.active_flows_origin.astype(int)))
                print("deltaflows = ", (saved_obs.active_flows_origin - obs.active_flows_origin).astype(int))
                print("final delta_flow = ", delta_flow)

                # 1) having the new and old obs. Now how do we compare ?
                simulated_score, worsened_line_ids, redistribution_prod, redistribution_load, efficacity = \
                    self.observations_comparator(saved_obs, obs, score_topo, delta_flow)

                if (self.isScoreFromBackend) and (simulated_score==4):
                    # dans le cas ou on resoud bien les contraintes, on prend la reward L2RPN
                    efficacity = reward

                # The next three lines have to been done like this to properly have a python empty list if no lines
                # are worsened. This is important to save, read back, and compare a DATAFRAME
                worsened_line_ids = list(np.where(worsened_line_ids == 1))
                worsened_line_ids = worsened_line_ids[0]

                # further tricks to properly read back a saved dataframe
                if worsened_line_ids.size == 0:
                    worsened_line_ids = []
                elif isinstance(worsened_line_ids, np.ndarray):
                    worsened_line_ids = list(worsened_line_ids)


                score_data = [self.ltc[0],
                              saved_obs.active_flows_origin[self.ltc[0]],
                              obs.active_flows_origin[self.ltc[0]],
                              delta_flow,
                              worsened_line_ids,
                              redistribution_prod,
                              redistribution_load,
                              new_conf,
                              new_conf,#the pypownet conf is the same as alphadeesp
                              self.external_to_internal_mapping[target_node],
                              1,  # category hubs?
                              score_topo,
                              simulated_score,
                              efficacity]

                max_index = end_result_dataframe.shape[0]  # rows
                end_result_dataframe.loc[max_index] = score_data

                end_result_dataframe.to_csv("./END_RESULT_DATAFRAME.csv", index=True)
                ii += 1
                j += 1

        return end_result_dataframe

    def observations_comparator(self, old_obs, new_obs, score_topo, delta_flow):
        """This function takes two observations and extracts several information:
        - the flow reports in %
        - list of lines, on which the situation got worse ie, the line capacity diminished.
            easily accessible with => obs.get_lines_capacity_usage()
        - prod volume diff
        - cons volume diff
        - score_simule => categories => 0, 1, 2, 3, 4
        - Efficacity =>

        it returns an array with all data for end_result_dataframe creation
        """
        simulated_score = self.score_changes_between_two_observations(old_obs, new_obs,self.environment.game.n_timesteps_actionned_line_reactionable)

        worsened_line_ids = self.create_boolean_array_of_worsened_line_ids(old_obs, new_obs)

        redistribution_prod = np.sum(np.absolute(new_obs.active_productions - old_obs.active_productions))
        redistribution_load = np.sum(np.absolute(new_obs.active_loads - old_obs.active_loads))

        if simulated_score in [4, 3, 2]:  # success
            efficacity = fabs(delta_flow / new_obs.get_lines_capacity_usage()[self.ltc[0]])
            pass
        elif simulated_score in [0, 1]:  # failure
            efficacity = -fabs(delta_flow / new_obs.get_lines_capacity_usage()[self.ltc[0]])
            pass
        else:
            raise ValueError("Cannot compute efficacity, the score is wrong.")

        return simulated_score, worsened_line_ids, redistribution_prod, redistribution_load, efficacity

    def create_boolean_array_of_worsened_line_ids(self, old_obs, new_obs):
        """This function creates a boolean array of lines that got worse between two observations.
        @:return boolean numpy array [0..1]"""

        res = []
        n_lines=len(new_obs.get_lines_capacity_usage())
        n_timesteps_actionned_line_reactionable=self.environment.game.n_timesteps_actionned_line_reactionable

        old_rho=old_obs.get_lines_capacity_usage()
        new_rho=new_obs.get_lines_capacity_usage()

        old_time_reco=old_obs.timesteps_before_lines_reconnectable
        new_time_reco=new_obs.timesteps_before_lines_reconnectable
        #for old, new in zip(old_obs, new_obs):
        for l in range(n_lines):
            if fabs(new_rho[l]) > 1 and fabs(old_rho[l]) > 1 and fabs(new_rho[l]) > 1.05 * fabs(old_rho[l]):  # contrainte existante empiree
                res.append(1)
            elif fabs(new_rho[l]) > 1 > fabs(old_rho[l]):
                res.append(1)
            elif (new_time_reco[l] - old_time_reco[l]>n_timesteps_actionned_line_reactionable):
                res.append(1)
            else:
                res.append(0)

        return np.array(res)

    def score_changes_between_two_observations(self, old_obs, new_obs,nb_timestep_cooldown_line_param=0):
        """This function takes two observations and computes a score to quantify the change between old_obs and new_obs.
        @:return int between [0 and 4]
        4: if every overload disappeared
        3: if an overload disappeared without stressing the network
        2: if at least 30% of an overload was relieved
        1: if an overload was relieved but another appeared and got worse
        0: if no overloads were alleviated or if it resulted in some load shedding or production distribution.
        """
        old_number_of_overloads = 0
        new_number_of_overloads = 0
        boolean_overload_worsened = []
        boolean_constraint_30percent_relieved = []
        boolean_constraint_relieved = []
        boolean_overload_created = []
        boolean_line_cascading_disconnection = ((new_obs.timesteps_before_lines_reconnectable - old_obs.timesteps_before_lines_reconnectable) > nb_timestep_cooldown_line_param)

        old_obs_lines_capacity_usage = old_obs.get_lines_capacity_usage()
        new_obs_lines_capacity_usage = new_obs.get_lines_capacity_usage()
        # ################################### PREPROCESSING #####################################
        for elem in old_obs_lines_capacity_usage:
            if elem > 1.0:
                old_number_of_overloads += 1

        for elem in new_obs_lines_capacity_usage:
            if elem > 1.0:
                new_number_of_overloads += 1

        # preprocessing for score 3 and 2
        line_id = 0
        for old, new in zip(old_obs_lines_capacity_usage, new_obs_lines_capacity_usage):
            # preprocessing for score 3
            if (new > 1.05 * old) & (new > 1.0):  # if new > old > 1.0 it means it worsened an existing constraint
                boolean_overload_worsened.append(1)
            else:
                boolean_overload_worsened.append(0)

            # preprocessing for score 2
            if (old > 1.0) & (line_id in self.ltc):  # if old was an overload:
                surcharge = old - 1.0
                diff = old - new
                percentage_relieved = diff * 100 / surcharge
                if percentage_relieved > 30.0:
                    boolean_constraint_30percent_relieved.append(1)
                else:
                    boolean_constraint_30percent_relieved.append(0)
            else:
                boolean_constraint_30percent_relieved.append(0)

            # preprocessing for score 1
            if (old > 1.0 > new) & (line_id in self.ltc):
                boolean_constraint_relieved.append(1)
            else:
                boolean_constraint_relieved.append(0)

            if old < 1.0 < new:
                boolean_overload_created.append(1)
            else:
                boolean_overload_created.append(0)

            line_id += 1

        boolean_overload_worsened = np.array(boolean_overload_worsened)
        boolean_constraint_30percent_relieved = np.array(boolean_constraint_30percent_relieved)
        boolean_constraint_relieved = np.array(boolean_constraint_relieved)
        boolean_overload_created = np.array(boolean_overload_created)

        redistribution_prod = np.sum(np.absolute(new_obs.active_productions - old_obs.active_productions))

        cut_load_percent = np.sum(np.absolute(new_obs.active_loads - old_obs.active_loads))/np.sum(old_obs.active_loads)

        # ################################ END OF PREPROCESSING #################################
        # score 0 if no overloads were alleviated or if it resulted in some load shedding or production distribution.
        if old_number_of_overloads == 0:
            # print("return NaN: No overflow at initial state of grid")
            return float('nan')
        elif cut_load_percent > 0.01:  # (boolean_overload_relieved == 0).all()
            # print("return 0: no overloads were alleviated or some load shedding occured.")
            return 0

        # score 1 if overload was relieved but another one appeared and got worse
        elif (boolean_constraint_relieved == 1).any() and ((boolean_overload_created == 1).any() or
                                                           (boolean_overload_worsened == 1).any() or (boolean_line_cascading_disconnection).any() ):
            # print("return 1: our overload was relieved but another one appeared")
            return 1

        # 4: if every overload disappeared
        elif old_number_of_overloads > 0 and new_number_of_overloads == 0:
            # print("return 4: every overload disappeared")
            return 4

        # 3: if this overload disappeared without stressing the network, ie,
        # if an overload disappeared
        # and without worsening existing constraint
        # and no Loads that get cut
        elif (boolean_constraint_relieved == 1).any() and \
                (boolean_overload_worsened == 0).all():
            # and \ (new_obs.are_loads_cut == 0).all():
            # print("return 3: our overload disappeared without stressing the network")
            return 3

        # 2: if at least 30% of this overload was relieved
        elif (boolean_constraint_30percent_relieved == 1).any() and \
                (boolean_overload_worsened == 0).all():
            # print("return 2: at least 30% of our overload [{}] was relieved".format(
            #     np.where(boolean_overload_30percent_relieved == 1)[0]))
            return 2

        # score 0
        elif (boolean_constraint_30percent_relieved == 0).all() or \
                (boolean_overload_worsened == 1).any():
            return 0

        else:
            raise ValueError("Probleme with Scoring")

    def create_and_fill_internal_structures(self, obs, df):
        """This function fills multiple structures:
        self.substation_elements, self.substation_to_node_mapping, self.internal_to_external_mapping
        @:arg observation, df"""
        # ################ PART I : fill self.internal_to_external_mapping
        substations_list = list(obs.substations_ids.astype(int))
        # we create mapping from external representation to internal.
        for i, substation_id in enumerate(substations_list):
            self.internal_to_external_mapping[i] = substation_id

        if self.internal_to_external_mapping:
            self.external_to_internal_mapping = self.invert_dict_keys_values(self.internal_to_external_mapping)

        # prod values
        prod_nodes = obs.productions_substations_ids
        cons_nodes = obs.loads_substations_ids
        prod_values = obs.active_productions
        cons_values = obs.active_loads

        # ################ PART II : fill self.substation_elements
        for substation_id in self.internal_to_external_mapping.keys():
            elements_array = []
            external_substation_id = self.internal_to_external_mapping[substation_id]
            current_conf, types = obs.get_nodes_of_substation(external_substation_id)

            if self.debug:
                print("--------------- substation_id {}", substation_id)
                print(current_conf)
                print(types)
                print(types[0])

            # here we create arrays of substations_ids indicating the destination for OriginLine
            indexes_tmp_or = np.where(obs.lines_or_substations_ids == external_substation_id)
            indexes_dest_or = [obs.lines_ex_substations_ids[ind] for ind in indexes_tmp_or]
            # this iterator contains a LIST OF SUBSTATION_IDS, which correspond to "the other end" of element line OR
            iter_indexes_dest_or = iter(indexes_dest_or[0])

            # here we create arrays of substations_ids indicating the destination for ExtremityLine
            indexes_tmp_ex = np.where(obs.lines_ex_substations_ids == external_substation_id)
            indexes_dest_ex = [obs.lines_or_substations_ids[ind] for ind in indexes_tmp_ex]
            # this iterator contains a LIST OF SUBSTATION_IDS, which correspond to "the other end" of element line EX
            iter_indexes_dest_ex = iter(indexes_dest_ex[0])

            assert (len(current_conf) == len(types))

            for busbar, elem in zip(current_conf, types):
                if elem == ElementType.PRODUCTION:
                    value = None
                    if external_substation_id in prod_nodes:
                        res_index = np.where(prod_nodes == external_substation_id)
                        value = prod_values[res_index]
                    elements_array.append(Production(busbar, value))

                elif elem == ElementType.CONSUMPTION:
                    value = None
                    if external_substation_id in cons_nodes:
                        res_index = np.where(cons_nodes == external_substation_id)
                        value = cons_values[res_index]
                    elements_array.append(Consumption(busbar, value))

                elif elem == ElementType.ORIGIN_POWER_LINE:
                    dest = self.external_to_internal_mapping[int(next(iter_indexes_dest_or))]
                    flow_value = list(df.query("idx_or == " + str(substation_id) + " & idx_ex == " + str(dest))
                                      ["delta_flows"].round(decimals=2))
                    if flow_value:  # if not empty
                        elements_array.append(OriginLine(busbar, dest, flow_value))
                    else:  # else means the flow has been swapped. We must invert edge.
                        flow_value = list(df.query("idx_ex == " + str(substation_id) + " & idx_or == " + str(dest))
                                          ["delta_flows"].round(decimals=2))
                        swapped_condition = \
                            list(df.query("idx_ex == " + str(substation_id) + " & idx_or == " + str(dest))
                                 ["swapped"])[0]
                        # second swapped_condition for new_flows_swapped in self.topo
                        second_condition = \
                            list(df.query("idx_ex == " + str(substation_id) + " & idx_or == " + str(dest))
                                 ["new_flows_swapped"])[0]

                        # if both are true, two swaps = do nothing or both are false and we do nothing.
                        if (swapped_condition and second_condition) or (not swapped_condition and not second_condition):
                            elements_array.append(OriginLine(busbar, dest, flow_value))

                        # if one condition is true
                        elif swapped_condition or second_condition:
                            elements_array.append(ExtremityLine(busbar, dest, flow_value))

                        else:
                            raise ValueError("Problem with swap conditions")

                elif elem == ElementType.EXTREMITY_POWER_LINE:
                    dest = self.external_to_internal_mapping[int(next(iter_indexes_dest_ex))]
                    flow_value = list(df.query("idx_or == " + str(dest) + " & idx_ex == " + str(substation_id))
                                      ["delta_flows"].round(decimals=2))

                    if flow_value:  # if not empty
                        elements_array.append(ExtremityLine(busbar, dest, flow_value))
                    else:
                        flow_value = list(df.query("idx_ex == " + str(dest) + " & idx_or == " + str(substation_id))
                                          ["delta_flows"].round(decimals=2))

                        swapped_condition = \
                            list(df.query("idx_ex == " + str(dest) + " & idx_or == " + str(substation_id))
                                 ["swapped"])[0]
                        # second swapped_condition for new_flows_swapped in self.topo
                        second_condition = \
                            list(df.query("idx_ex == " + str(dest) + " & idx_or == " + str(substation_id))
                                 ["new_flows_swapped"])[0]

                        # if both are true, two swaps = do nothing or both are false and we do nothing.
                        if (swapped_condition and second_condition) or (not swapped_condition and not second_condition):
                            elements_array.append(ExtremityLine(busbar, dest, flow_value))
                        # if one condition is true
                        elif swapped_condition or second_condition:
                            elements_array.append(OriginLine(busbar, dest, flow_value))
                        else:
                            raise ValueError("Problem with swap conditions")

            self.substations_elements[substation_id] = elements_array

    def load(self):
        self.load_from_observation(self.obs, self.ltc)

    def load_from_observation(self, observation, lines_to_cut: list):
        # first, load information into a data frame
        self.ltc = lines_to_cut
        # d is a dict containing topology
        d = self.extract_topo_from_obs(observation)
        self.topo = d
        df = self.create_df(d, lines_to_cut)
        self.df = df
        print("DF From load2")
        print(df)
        self.create_and_fill_internal_structures(observation, df)


    def build_detailed_graph_from_internal_structure(self, lines_to_cut):
        """This function create a detailed graph from internal self structures as self.substations_elements..."""
        g = nx.MultiDiGraph()
        network = Network(self.substations_elements)
        print("Network = ", network)
        build_nodes_v2(g, network.nodes_prod_values)
        build_edges_v2(g, network.substation_id_busbar_id_node_id_mapping, self.substations_elements)
        print("This graph is weakly connected : ", nx.is_weakly_connected(g))
        return g

    def change_nodes_configurations(self, new_configuration, node_id):
        """Changes pypownet's internal graph network by changing node : node_id by applying new_configuration"""

        action_space = self.environment.action_space
        action = action_space.get_do_nothing_action(as_class_Action=True)

        for new_conf, id_node in zip(new_configuration, node_id):
            action_space.set_substation_switches_in_action(action, id_node, new_conf)
        raw_simulated_obs, *_ = self.environment.simulate(action)
        # if obs is None, error in the simulation of the next step
        if raw_simulated_obs is None:
            raise ValueError("\n\nPypownet simulation returned a None... Cannot process results...\n")
        # transform raw_obs into Observation object. (Useful for prints, for debugging)
        obs = self.environment.observation_space.array_to_observation(raw_simulated_obs)
        return obs

    @staticmethod
    def extract_topo_from_obs(obs):
        """This function, takes an obs an returns a dict with all topology information"""
        d = {
            "edges": {},
            "nodes": {}
        }
        # print("lines_cut = ", lines_cut)
        nodes_ids = obs.substations_ids
        # print("obs substations_ids = ", nodes_ids)
        idx_or = [int(x - 1) for x in obs.lines_or_substations_ids]
        idx_ex = [int(x - 1) for x in obs.lines_ex_substations_ids]
        prods_ids = obs.productions_substations_ids
        loads_ids = obs.loads_substations_ids
        are_prods = [node_id in prods_ids for node_id in nodes_ids]
        are_loads = [node_id in loads_ids for node_id in nodes_ids]
        prods_values = obs.active_productions
        loads_values = obs.active_loads
        current_flows = obs.active_flows_origin
        d["edges"]["idx_or"] = [x for x in idx_or]
        d["edges"]["idx_ex"] = [x for x in idx_ex]
        d["edges"]["init_flows"] = current_flows
        d["nodes"]["are_prods"] = are_prods
        d["nodes"]["are_loads"] = are_loads
        d["nodes"]["prods_values"] = prods_values
        d["nodes"]["loads_values"] = loads_values
        return d

    def cut_lines_and_recomputes_flows(self, ids: list):
        """This functions cuts lines: [ids], simulates and returns new line flows"""
        action_space = self.environment.action_space
        action = action_space.get_do_nothing_action(as_class_Action=True)
        for line_id in ids:
            action_space.set_lines_status_switch_from_id(action=action, line_id=line_id, new_switch_value=1)
        raw_simulated_obs = self.environment.simulate(action)
        if raw_simulated_obs[0] is None:
            raise ValueError("The simulation step of Pypownet returnt a None... Something")
        obs = self.environment.observation_space.array_to_observation(raw_simulated_obs[0])
        self.obs_linecut = obs
        self.lines_outaged_cut = np.argwhere(self.obs_linecut.lines_status == 0)

        return obs.active_flows_origin


def build_nodes_v2(g, nodes_prod_values: list):
    """nodes_prod_values is a list of tuples, (graphical_node_id, prod_cons_total_value)
        prod_cons_total_value is a float.
        If the value is positive then it is a Production, if negative it is a Consumption
    """
    print("IN FUNCTION BUILD NODES V2222222222", nodes_prod_values)
    for data in nodes_prod_values:
        print("data = ", data)
        i = int(data[0])
        if data[1] is None or data[1] == "XXX":
            prod_minus_load = 0.0  # It will end up as a white node
        else:
            prod_minus_load = data[1]
        print("prod_minus_load = ", prod_minus_load)
        if prod_minus_load > 0:  # PROD
            g.add_node(i, pin=True, prod_or_load="prod", value=str(prod_minus_load), style="filled",
                       fillcolor="#f30000")  # red color
        elif prod_minus_load < 0:  # LOAD
            g.add_node(i, pin=True, prod_or_load="load", value=str(prod_minus_load), style="filled",
                       fillcolor="#478fd0")  # blue color
        else:  # WHITE COLOR
            g.add_node(i, pin=True, prod_or_load="load", value=str(prod_minus_load), style="filled",
                       fillcolor="#ffffed")  # white color
        i += 1


def build_edges_v2(g, substation_id_busbar_id_node_id_mapping, substations_elements):
    print("\nWE ARE IN BUILD EDGES V2")
    substation_ids = sorted(list(substations_elements.keys()))
    # loops through each substation, and creates an edge from (
    for substation_id in substation_ids:
        print("\nSUBSTATION ID = ", substation_id)
        for element in substations_elements[substation_id]:
            print(element)
            origin = None
            extremity = None
            if isinstance(element, OriginLine):
                # origin = substation_id
                origin = int(substation_id_busbar_id_node_id_mapping[substation_id][element.busbar_id])
                extremity = int(element.end_substation_id)
                # check if extremity on busbar1, if it is,
                # check with the substation substation_id_busbar_id_node_id_mapping dic what "graphical" node it is
                print("substations_elements[extremity] = ", substations_elements[extremity])
                for elem in substations_elements[extremity]:
                    # if this true, we are talking about correct edge
                    if isinstance(elem, ExtremityLine) and elem.flow_value == element.flow_value:
                        if elem.busbar == 1:
                            extremity = substation_id_busbar_id_node_id_mapping[extremity][1]
                reported_flow = element.flow_value
            elif origin is None or extremity is None:
                continue
            # in case we get on an element that is Production or Consumption
            else:
                continue
            print("origin = ", origin)
            print("extremity = ", extremity)
            print("reported_flow = ", reported_flow)
            pen_width = fabs(reported_flow[0]) / 10.0
            if pen_width < 0.01:
                pen_width = 0.1
            print(f"#################### Edge created : ({origin}, {extremity}), with flow = {reported_flow},"
                  f" pen_width = {pen_width} >>>")
            if reported_flow[0] > 0:  # RED
                g.add_edge(origin, extremity, capacity=float(reported_flow[0]), label=reported_flow[0], color="coral",
                           penwidth="%.2f" % pen_width)
            else:  # BLUE
                g.add_edge(origin, extremity, capacity=float(reported_flow[0]), label=reported_flow[0], color="blue",
                           penwidth="%.2f" % pen_width)
            g.add_edge(origin, extremity, capacity=float(reported_flow[0]), label=reported_flow[0])


def get_differencial_topology(new_conf, old_conf):
    """new - old, for elem in result, if elem -1, then put one"""
    assert (len(new_conf) == len(old_conf))
    res = []

    for elemNew, elemOld in zip(new_conf, old_conf):
        r = elemNew - elemOld
        if r < 0:
            res.append(1)
        else:
            res.append(int(r))
    return res



from grid2op.Action import TopologyAction
from grid2op.Reward import L2RPNReward
from grid2op.Rules import DefaultRules
from grid2op.Chronics import Multifolder
from grid2op.Backend import PandaPowerBackend

config = {
    "backend": PandaPowerBackend,
    "action_class": TopologyAction,
    "observation_class": None,
    "reward_class": L2RPNReward,
    "gamerules_class": DefaultRules,
    "chronics_class": Multifolder,
    "volagecontroler_class": None,
    "thermal_limits": [1740, 500, 500, 500, 500, 500, 500, 500, 500, 230, 500, 500,
                       500, 500, 500, 500, 500, 500, 500, 500],
}
